"""
utilities to apply frequencies
"""
import numpy as np
import pandas as pd
from enum import Enum
from typing import Optional, Union, Callable, Literal

import qis.utils.dates as da
from qis.utils.df_str import df_index_to_str

# Arguments for fillna()
FillnaOptions = Literal["bfill", "ffill", "pad"]


def df_asfreq(df: Union[pd.DataFrame, pd.Series],
              freq: Optional[str] = 'QE',
              method: FillnaOptions = 'ffill',
              fill_na_method: Optional[FillnaOptions] = 'ffill',
              inclusive: Optional[str] = None,
              include_start_date: bool = False,
              include_end_date: bool = False,
              tz: str = None
              ) -> Union[pd.DataFrame, pd.Series]:
    """
    wrapper to asfreq with closed period
    closed{None, ‘left’, ‘right’}, optional : Make the interval closed with respect to the given frequency
    to the ‘left’, ‘right’, or both sides (None, the default).
    note that using include_start_date and include_end_date will disable given index freq
    """
    if freq is None:
        return df
    elif len(df.index) >=3 and freq == pd.infer_freq(df.index):
        return df

    freq_index = da.generate_dates_schedule(time_period=da.get_time_period(df=df, tz=tz), freq=freq,
                                            include_start_date=include_start_date,
                                            include_end_date=include_end_date)
    if freq_index.empty:
        print(f"in df_asfreq: cannot resample with freq={freq} with start={df.index[0]}, end={df.index[-1]} - using start/end")
        freq_index = freq_index.insert(0, df.index[0])
        freq_index = freq_index.insert(1, df.index[-1])
    if include_start_date:
        if freq_index[0] != df.index[0]:
            freq_index = freq_index.insert(0, df.index[0])
    if include_end_date:  # close on last date
        if freq_index[-1] != df.index[-1]:
            freq_index = freq_index.append(df.index[-1:])
    freq_data = df.reindex(index=freq_index, method=method)

    if fill_na_method is not None:
        if fill_na_method == 'ffill':
            freq_data = freq_data.ffill()
        elif fill_na_method == 'bfill':
            freq_data = freq_data.bfill()
        else:
            raise NotImplementedError(f"fill_na_method={fill_na_method}")
    return freq_data


def agg_remained_data_on_right(df: Union[pd.DataFrame, pd.Series],
                               data: Union[pd.DataFrame, pd.Series],
                               agg_func: Optional[Callable[[pd.DataFrame], pd.Series]]  # for none use last
                               ) -> Union[pd.DataFrame, pd.Series]:
    if df.index[-1] < data.index[-1]:  # some data left on right
        remained_data_on_right = data.loc[df.index[-1]:]
        if df.index[-1] in remained_data_on_right.index:  # data_f.index[-1] can be included in previous sample
            remained_data_on_right = remained_data_on_right.drop(df.index[-1])
        if agg_func is not None:
            agg_remained_data_on_right = remained_data_on_right.apply(agg_func)
        else:
            agg_remained_data_on_right = remained_data_on_right.iloc[-1]
        df = df.append(agg_remained_data_on_right)
    else:
        df = df
    return df


def df_resample_at_other_index(df: Union[pd.DataFrame, pd.Series],
                               other_index: Union[pd.DatetimeIndex, pd.Index],
                               agg_func: Callable[[pd.DataFrame], pd.Series] = np.nanmean,
                               fill_na_method: FillnaOptions = 'ffill',
                               include_end_date: bool = False
                               ) -> Union[pd.DataFrame, pd.Series]:
    """
    given the time index of another time series, aggregate data at frequency of the index
    """
    if not isinstance(other_index, pd.DatetimeIndex):
        raise TypeError (f"other_index type = {type(other_index)} must be pd.DatetimeIndex")

    freq = pd.infer_freq(other_index)
    if freq is None:
        raise ValueError(f"could not infer frequency for index = {other_index}")

    data_f = df.resample(freq).apply(agg_func)

    if include_end_date:
        data_f = agg_remained_data_on_right(df=data_f, data=df, agg_func=agg_func)

    if fill_na_method is not None:
        if fill_na_method == 'ffill':
            data_f = data_f.reindex(index=other_index).ffill()
        elif fill_na_method == 'bfill':
            data_f = data_f.reindex(index=other_index).bfill()
        else:
            raise NotImplementedError(f"fill_na_method={fill_na_method}")

    return data_f


def df_resample_at_freq(df: Union[pd.DataFrame, pd.Series],
                        freq: str = 'QE',
                        fill_na_method: FillnaOptions = 'ffill',
                        agg_func: Optional[Callable[[pd.DataFrame], pd.Series]] = np.nanmean,  # if none use last
                        include_end_date: bool = False
                        ) -> Union[pd.DataFrame, pd.Series]:
    """
    wrapper to asfreq with closed period
    problem with resample it can put last date above last date in data
    """
    # other index date is strictly below the last date in data
    insample_index = pd.date_range(start=df.index[0], end=df.index[-1], freq=freq)
    in_sample_data = df.loc[:insample_index[-1]]
    if agg_func is not None:
        data_f = in_sample_data.resample(freq).apply(agg_func)
    else:
        data_f = in_sample_data.resample(freq).last()

    if include_end_date:
        data_f = agg_remained_data_on_right(df=data_f, data=df, agg_func=agg_func)

    if fill_na_method is not None:  # final check
        if fill_na_method == 'ffill':
            data_f = data_f.ffill()
        elif fill_na_method == 'bfill':
            data_f = data_f.bfill()
        else:
            raise NotImplementedError(f"fill_na_method={fill_na_method}")

    return data_f


def df_resample_at_int_index(df: pd.DataFrame,
                             func: Callable = np.nansum,
                             sample_size: int = 5
                             ) -> pd.DataFrame:
    """
    resample dataframe at evenly spaced discrete index
    """
    assert isinstance(sample_size, int)
    original_index = df.index
    df = df.reset_index(drop=True)  # need integer range index
    sampler = (df.index.to_series() / sample_size).astype(int)
    df = df.groupby(sampler).agg(func, axis=0)
    if isinstance(df, pd.Series):  # if input is one columns
        df = df.to_frame()

    # define sampled index
    sampled_index = original_index[sample_size-1::sample_size]
    if len(sampled_index) < len(df.index):
        sampled_index = sampled_index.append(pd.Index([original_index[-1]]))
    df = df.set_index(sampled_index)
    return df


class UnitTests(Enum):
    AS_FREQ = 1
    RESAMPLE = 2
    INT_INDEX = 3


def run_unit_test(unit_test: UnitTests):

    time_period = da.TimePeriod('1Jan2020', '1Dec2020')
    daily_index = time_period.to_pd_datetime_index(freq='B')
    df = pd.DataFrame(data=np.tile(np.array([1.0, 2.0]), (len(daily_index), 1)), index=daily_index, columns=['1', '2'])
    print(df)

    if unit_test == UnitTests.AS_FREQ:
        freq_data = df_asfreq(df=df, freq='YE')
        print(freq_data)
        freq_data = df_asfreq(df=df, freq='YE', include_end_date=True)
        print(freq_data)
        print(type(freq_data.index))

        freq_data_s = df_index_to_str(freq_data)
        print(freq_data_s)
        print(type(freq_data_s.index))

    elif unit_test == UnitTests.RESAMPLE:

        time_period1 = da.TimePeriod('1Jan2020', '1Jan2021')
        other_index = time_period1.to_pd_datetime_index(freq='QE')
        print(other_index)
        freq_data1 = df_resample_at_other_index(df=df, other_index=other_index, agg_func=np.nansum)
        print(freq_data1)
        print(freq_data1.index)

        freq_data2 = df_resample_at_freq(df=df, freq='QE', agg_func=np.nansum, include_end_date=True)
        print(freq_data2)
        print(freq_data2.index)

    elif unit_test == UnitTests.INT_INDEX:
        df = df_resample_at_int_index(df=df, sample_size=21)
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.INT_INDEX

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import qis as qis
from qis import PerfStat
from bbg_fetch import fetch_field_timeseries_per_tickers

assets = {'BASPCAAA Index': 'AAA10y',
          'BICLA10Y Index': 'A10y',
          'BICLB10Y Index': 'BAA10y',
          'CSI BARC Index': 'HY 10Y'}

spreads = fetch_field_timeseries_per_tickers(tickers=list(assets.keys()), field='PX_LAST', CshAdjNormal=True)
spreads = spreads.rename(assets, axis=1)
spreads['HY 10Y'] = 100.0*spreads['HY 10Y']
# to %
spreads = spreads / 10000.0

predictors = {'SPXT Index': 'SPX', 'gt10 Govt': 'UST10y'}
benchmarks = fetch_field_timeseries_per_tickers(tickers=list(predictors.keys()), field='PX_LAST', CshAdjNormal=True)
benchmarks = benchmarks.rename(predictors, axis=1)
print(benchmarks)
qis.plot_time_series(spreads,
                     var_format='{:,.2%}', x_date_freq='YE')

freq = 'ME'

df1 = pd.concat([benchmarks['SPX'].asfreq(freq, method='ffill').pct_change(),
                 spreads.asfreq(freq, method='ffill').diff()], axis=1).dropna()

qis.plot_scatter(df=df1, x='SPX')

df2 = pd.concat([benchmarks['UST10y'].asfreq(freq, method='ffill').diff()/100.0,
                 spreads.asfreq(freq, method='ffill').diff()], axis=1).dropna()

qis.plot_scatter(df=df2, x='UST10y')

plt.show()







"""
define transform for ra returns
"""
# packages
import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.stats import laplace
from typing import Union, Optional, Tuple
from enum import Enum
# qis
import qis.utils.dates as da
import qis.utils.np_ops as npo
import qis.models.linear.ewm as ewm


def compute_ra_returns(returns: Union[pd.Series, pd.DataFrame],
                       ewm_lambda: Union[float, np.ndarray] = 0.94,
                       vol_target: Optional[float] = 0.12,
                       span: Optional[int] = None,
                       weight_lag: Optional[int] = 1,
                       is_log_returns_to_arithmetic: bool = True  # typically log-return are passed to vol computations
                       ) -> Tuple[Union[pd.Series, pd.DataFrame], Union[pd.Series, pd.DataFrame], Union[pd.Series, pd.DataFrame]]:

    if span is not None:
        ewm_lambda = 1.0-2.0/(span+1.0)

    if vol_target is None:  # non-dimensional 100% vol target
        annualize = False
        vol_target = 1.0
    else:
        annualize = True

    ewm_vol = ewm.compute_ewm_vol(data=returns,
                                  ewm_lambda=ewm_lambda,
                                  mean_adj_type=ewm.MeanAdjType.NONE,
                                  annualize=annualize)

    weights = npo.to_finite_reciprocal(data=ewm_vol, fill_value=0.0, is_gt_zero=True)
    weights = weights.multiply(vol_target)
    if weight_lag is not None:
        weights = weights.shift(weight_lag)

    if is_log_returns_to_arithmetic:  # convert returns back to arithmetic = exp(r)-1.0
        returns = np.expm1(returns)

    ra_returns = returns.multiply(weights)

    # alignment in case
    if isinstance(ra_returns, pd.DataFrame):
        ra_returns = ra_returns[returns.columns]
        weights = weights[returns.columns]

    return ra_returns, weights, ewm_vol


def compute_ewm_long_short_filtered_ra_returns(returns: pd.DataFrame,
                                               vol_span: Optional[Union[int, np.ndarray]] = 31,
                                               long_span: Union[int, np.ndarray] = 63,
                                               short_span: Optional[Union[int, np.ndarray]] = 5,
                                               warmup_period: Optional[Union[int, np.ndarray]] = 21,
                                               weight_lag: Optional[int] = 1
                                               ) -> pd.DataFrame:
    if vol_span is not None:
        ra_returns, _, _ = compute_ra_returns(returns=returns, span=vol_span, vol_target=None, weight_lag=weight_lag)
    else:
        ra_returns = returns
    filter = ewm.compute_ewm_long_short_filter(data=ra_returns,
                                               long_span=long_span,
                                               short_span=short_span,
                                               warmup_period=warmup_period)
    return filter


class SignalMapType(Enum):
    NormalCDF = 1
    LaplaceCDF = 2
    ExpCDF = 3


def map_signal_to_weight(signals: pd.DataFrame, signal_map_type: SignalMapType = SignalMapType.NormalCDF,
                         loc: Union[float, pd.DataFrame] = 0.0,
                         scale: Union[float, np.ndarray] = 1.0,
                         tail_level: Union[float, np.ndarray] = 1.0,
                         slope_right: Union[float, np.ndarray] = 0.5,
                         slope_left: Union[float, np.ndarray] = 0.5,
                         tail_decay_right: Optional[Union[float, np.ndarray]] = None,
                         tail_decay_left: Optional[Union[float, np.ndarray]] = None
                         ) -> pd.DataFrame:
    x = signals.to_numpy()
    if isinstance(loc, pd.DataFrame):
        loc = loc.to_numpy()
    if isinstance(scale, np.ndarray) and scale.shape[0] != x.shape[1]:
        raise ValueError(f"{scale.shape[0]} != {x.shape[1]}")
    if isinstance(slope_right, np.ndarray) and slope_right.shape[0] != x.shape[1]:
        raise ValueError(f"{slope_right.shape[0]} != {x.shape[1]}")
    if isinstance(slope_left, np.ndarray) and slope_left.shape[0] != x.shape[1]:
        raise ValueError(f"{slope_left.shape[0]} != {x.shape[1]}")
    if isinstance(tail_level, np.ndarray) and tail_level.shape[0] != x.shape[1]:
        raise ValueError(f"{tail_level.shape[0]} != {x.shape[1]}")

    if signal_map_type == SignalMapType.NormalCDF:
        weight = 2.0*norm.cdf(x=x, loc=loc, scale=scale) - 1.0

    elif signal_map_type == SignalMapType.LaplaceCDF:
        weight = 2.0*laplace.cdf(x=x, loc=loc, scale=scale) - 1.0

    elif signal_map_type == SignalMapType.ExpCDF:
        if np.any(np.less_equal(tail_level, slope_right)) or np.any(np.less_equal(tail_level, slope_left)):
            raise ValueError(f"must be tail>slope_positive and tail > slope_negative")
        scale_negative = 1.5625 * scale / np.log(tail_level / (tail_level - slope_left))
        scale_positive = 1.5625 * scale / np.log(tail_level / (tail_level - slope_right))
        s_negative = - tail_level * (1.0 - np.exp(-np.square(x - loc) / scale_negative))
        s_positive = tail_level * (1.0 - np.exp(-np.square(x - loc) / scale_positive))
        weight = np.where(np.less(x, loc, where=np.isfinite(x)), s_negative, s_positive)

        if tail_decay_right is not None and tail_decay_left is not None:  # take min(loc,0.0) and max(loc, 0.0)
            if isinstance(tail_decay_right, np.ndarray) and tail_decay_right.shape[0] != x.shape[1]:
                raise ValueError(f"{tail_decay_right.shape[0]} != {x.shape[1]}")
            if isinstance(tail_decay_left, np.ndarray) and tail_decay_left.shape[0] != x.shape[1]:
                raise ValueError(f"{tail_decay_left.shape[0]} != {x.shape[1]}")

            x_left_tail = x + tail_level - np.where(np.less(loc, 0.0), loc, 0.0)
            f_left_tail = np.where(np.less(x_left_tail, 0.0), np.exp(x_left_tail / tail_decay_left), 1.0)

            x_right_tail = x - tail_level - np.where(np.greater(loc, 0.0), loc, 0.0)
            f_right_tail = np.where(np.greater(x_right_tail, 0.0), np.exp(-x_right_tail / tail_decay_right), 1.0)

            tails = np.where(np.greater(x_right_tail, 0.0), f_right_tail, f_left_tail)
            weight = weight * tails

    else:
        raise NotImplementedError(f"signal_map_type={signal_map_type}")
    weight = pd.DataFrame(weight, index=signals.index, columns=signals.columns)
    return weight


def compute_rolling_ra_returns(returns: pd.DataFrame,
                               span: int = 1,
                               ewm_lambda_eod: float = 0.94,
                               vol_target: Optional[float] = None,
                               weight_shift: Optional[int] = 1,
                               is_log_returns_to_arithmetic: bool = True  # typically log-return are passed to vol
                               ) -> pd.DataFrame:
    """
    span = 1: daily ra returns
    otherwise compute sum of returns and then their vols
    interpretation: voltargeting for returns over span
    ewm_lambda is vol over the span too
    """
    if span > 1:
        rolling_returns = returns.rolling(span).sum()
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    else:
        ewm_lambda = ewm_lambda_eod
        rolling_returns = returns

    ra_returns, weights, _ = compute_ra_returns(returns=rolling_returns,
                                                ewm_lambda=ewm_lambda,
                                                vol_target=vol_target,
                                                weight_lag=weight_shift,
                                                is_log_returns_to_arithmetic=is_log_returns_to_arithmetic)
    return ra_returns


def compute_sum_rolling_ra_returns(returns: pd.DataFrame,
                                   span: int = 1,
                                   ewm_lambda: float = 0.94,
                                   vol_target: Optional[float] = None,
                                   weight_shift: Optional[int] = 1,
                                   is_log_returns_to_arithmetic: bool = True,  # typically log-return are passed to vol
                                   is_norm: bool = True
                                   ) -> pd.DataFrame:
    """
    span = 1: daily ra returns
    otherwise compute sum of daily ra-returns
    interpretation: pnl of daily voltargeting returns over span
    """
    ra_returns, weights, _ = compute_ra_returns(returns=returns,
                                                ewm_lambda=ewm_lambda,
                                                vol_target=vol_target,
                                                weight_lag=weight_shift,
                                                is_log_returns_to_arithmetic=is_log_returns_to_arithmetic)

    if span > 1:
        sum_rolling_ra_returns = ra_returns.rolling(span).sum()

        if is_norm:
            sum_rolling_ra_returns = sum_rolling_ra_returns.divide(np.sqrt(span))
    else:
        sum_rolling_ra_returns = ra_returns

    return sum_rolling_ra_returns


def compute_sum_freq_ra_returns(returns: Union[pd.Series, pd.DataFrame],
                                freq: str = 'B',
                                ewm_lambda: float = 0.94,
                                vol_target: Optional[float] = None,
                                weight_shift: Optional[int] = 1,
                                is_log_returns_to_arithmetic: bool = True,  # typically log-return are passed to vol
                                is_norm: bool = True
                                ) -> Union[pd.Series, pd.DataFrame]:
    """
    span = 1: daily ra returns
    otherwise compute freq sum of daily ra-returns
    interpretation: pnl of daily voltargeting returns over non-overlap freq
    """
    ra_returns, weights, _ = compute_ra_returns(returns=returns,
                                                ewm_lambda=ewm_lambda,
                                                vol_target=vol_target,
                                                weight_lag=weight_shift,
                                                is_log_returns_to_arithmetic=is_log_returns_to_arithmetic)

    if freq not in ['B', 'D']:
        sum_rolling_ra_returns = ra_returns.resample(freq).sum()

        if is_norm:
            n_daily, _ = da.get_period_days(freq=freq, is_calendar=False)
            sum_rolling_ra_returns = sum_rolling_ra_returns.divide(np.sqrt(n_daily))
    else:
        sum_rolling_ra_returns = ra_returns

    return sum_rolling_ra_returns


def compute_ewm_ra_returns_momentum(returns: Union[pd.Series, pd.DataFrame],
                                    momentum_span: int = 63,
                                    momentum_lambda: Optional[Union[float, np.ndarray]] = None,
                                    vol_span: int = 31,
                                    vol_lambda: Optional[Union[float, np.ndarray]] = None,
                                    weight_shift: Optional[int] = 1
                                    ) -> Union[pd.Series, pd.DataFrame]:
    """
    span = 1: daily ra returns
    """
    if momentum_lambda is None:
        momentum_lambda = 1.0 - 2.0 / (momentum_span + 1.0)
    if vol_lambda is None:
        vol_lambda = 1.0 - 2.0 / (vol_span + 1.0)

    ra_returns, _, _ = compute_ra_returns(returns=returns,
                                          ewm_lambda=vol_lambda,
                                          vol_target=None,
                                          weight_lag=weight_shift)

    ewm_signal = ewm.ewm_recursion(a=ra_returns.to_numpy(),
                                   ewm_lambda=momentum_lambda,
                                   init_value=0.0 if isinstance(returns, pd.Series) else np.zeros(len(returns.columns)),
                                   is_unit_vol_scaling=True)

    if isinstance(returns, pd.DataFrame):
        ewm_ra_returns_momentum = pd.DataFrame(data=ewm_signal, index=returns.index, columns=returns.columns)
    else:
        ewm_ra_returns_momentum = pd.Series(data=ewm_signal, index=returns.index, name=returns.name)

    return ewm_ra_returns_momentum


def get_paired_rareturns_signals(returns: Union[pd.Series, pd.DataFrame],
                                 signal: Union[pd.Series, pd.DataFrame],
                                 freq: str = 'BQ',
                                 span: int = 63,
                                 is_nonoverlapping: bool = True,
                                 ra_returns_ewm_vol_lambda: float = 0.94,
                                 is_mean_adjust_returns: bool = False
                                 ) -> Tuple[Union[pd.Series, pd.DataFrame], Union[pd.Series, pd.DataFrame]]:

    if is_nonoverlapping:
        sum_freq_ra_returns = compute_sum_freq_ra_returns(returns=returns,
                                                              freq=freq,
                                                              ewm_lambda=ra_returns_ewm_vol_lambda,
                                                              is_norm=True)
        ra_return = sum_freq_ra_returns
        indicator = signal.resample(freq).last().shift(1)
    else:
        sum_rolling_ra_returns = compute_sum_rolling_ra_returns(returns=returns,
                                                                    span=span,
                                                                    ewm_lambda=ra_returns_ewm_vol_lambda,
                                                                    is_norm=True)
        ra_return = sum_rolling_ra_returns
        indicator = signal.shift(1)

    # into two column dataframe
    if is_mean_adjust_returns:
        ra_returns_mean = ra_return.expanding(min_periods=1, axis=0).mean()  # apply pandas expanding
        ra_return = ra_return.subtract(ra_returns_mean)

    return ra_return, indicator


class ReturnsTransform(Enum):
    ROLLING_RA_RETURNS = 1
    EWMA_RETURNS_MOMENTUM = 2


def compute_returns_transform(returns: pd.DataFrame,
                              returns_transform: ReturnsTransform = ReturnsTransform.ROLLING_RA_RETURNS,
                              momentum_span: int = 31,
                              vol_span: int = 31,
                              rolling_ra_returns_span: int = 31
                              ) -> pd.DataFrame:

    if returns_transform == ReturnsTransform.ROLLING_RA_RETURNS:
        returns_transform = compute_rolling_ra_returns(returns=returns,
                                                       span=rolling_ra_returns_span,
                                                       weight_shift=1)
    elif returns_transform == ReturnsTransform.EWMA_RETURNS_MOMENTUM:
        returns_transform = compute_ewm_ra_returns_momentum(returns=returns,
                                                             momentum_span=momentum_span,
                                                             vol_span=vol_span,
                                                             weight_shift=1)
    else:
        raise TypeError(f"returns_transform {returns_transform} of {type(returns_transform)} not implemented")
    return returns_transform


class UnitTests(Enum):
    RA_RETURNS = 1
    TRANSFORM = 2

def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()
    returns = prices.pct_change()

    if unit_test == UnitTests.RA_RETURNS:
        df = compute_ra_returns(returns=returns)
        print(df)

    elif unit_test == UnitTests.TRANSFORM:
        df = compute_returns_transform(returns=returns)
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.RA_RETURNS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




"""
errorbar plot
"""

# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Union, Tuple, Optional
from enum import Enum

# qis
import qis.plots.utils as put


def plot_errorbar(df: Union[pd.Series, pd.DataFrame],
                  y_std_errors: Union[float, pd.Series, pd.DataFrame] = 0.5,
                  exact: pd.Series = None,  # can add exact solution
                  legend_title: str = None,
                  legend_loc: Optional[Union[str, bool]] = 'upper left',
                  xlabel: str = None,
                  ylabel: str = None,
                  var_format: Optional[str] = '{:.0f}',
                  title: Union[str, bool] = None,
                  fontsize: int = 10,
                  capsize: int = 10,
                  colors: List[str] = None,
                  exact_color: str = 'green',
                  exact_marker: str = "v",
                  y_limits: Tuple[Optional[float], Optional[float]] = None,
                  ax: plt.Subplot = None,
                  **kwargs
                  ) -> Optional[plt.Figure]:

    if isinstance(df, pd.DataFrame):
        pass
    elif isinstance(df, pd.Series):
        df = df.to_frame()
    else:
        raise TypeError(f"unsupported data type {type(df)}")

    columns = df.columns

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
        colors = put.get_n_colors(n=len(columns), **kwargs)

    for idx, column in enumerate(columns):
        if isinstance(y_std_errors, pd.DataFrame):
            yerr = y_std_errors[column].to_numpy()  # columnwese
        elif isinstance(y_std_errors, pd.Series):
            yerr = y_std_errors.to_numpy()
        else:
            yerr = y_std_errors

        ax.errorbar(x=df.index, y=df[column].to_numpy(), yerr=yerr, color=colors[idx], fmt='o', capsize=capsize)

    if exact is not None:
        for idx, index in enumerate(df.index):
            put.add_scatter_points(ax=ax, label_x_y=[(index, exact[index])], color=exact_color,
                                   marker=exact_marker, **kwargs)
        labels = columns.to_list() + [exact.name]
        colors = colors + [exact_color]
        markers = ['o']*len(columns) + [exact_marker]
    else:
        labels = columns
        markers = ['o']*len(columns)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    if legend_loc is not None:
        put.set_legend(ax=ax,
                       markers=markers,
                       labels=labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       legend_title=legend_title,
                       handlelength=0,
                       fontsize=fontsize,
                       **kwargs)

    else:
        ax.legend().set_visible(False)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    if var_format is not None:
        put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format, **kwargs)
    else:
        put.set_ax_ticks_format(ax=ax, fontsize=fontsize, **kwargs)
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_spines(ax=ax, **kwargs)

    return fig


class UnitTests(Enum):
    ERROR_BAR = 1


def run_unit_test(unit_test: UnitTests):

    n = 10
    x = np.linspace(0, 10, n)
    dy = 0.8
    y1 = pd.Series(np.sin(x) + dy * np.random.randn(n), index=x, name='y1')
    y2 = pd.Series(np.cos(x) + dy * np.random.randn(n), index=x, name='y2')
    data = pd.concat([y1, y2], axis=1)

    if unit_test == UnitTests.ERROR_BAR:

        global_kwargs = {'fontsize': 8,
                         'linewidth': 0.5,
                         'weight': 'normal',
                         'markersize': 1}

        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(8, 6))
            plot_errorbar(df=data,
                          ax=ax,
                          **global_kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.ERROR_BAR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
plot aggregate report on pandas data and plot time series of columns
each time series plot = data[column]
 # = # columns
"""
# packages
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Tuple, Optional, Dict, List
from enum import Enum

# qis
import qis as qis
import qis.utils.dates as da
import qis.utils.df_agg as dfa
import qis.utils.df_groups as dfg
import qis.utils.df_ops as dfo
from qis.perfstats.desc_table import DescTableType
from qis.plots import histogram as pdf
from qis.plots.reports.utils import get_summary_table_fig, set_x_date_freq, ReportType
from qis.plots.derived.data_timeseries import plot_data_timeseries


def econ_data_report(data: pd.DataFrame,
                     complimentary_datas: Dict[str, pd.DataFrame] = None,
                     group_data: pd.Series = None,
                     descriptive_data: pd.DataFrame = None,
                     inst_names: pd.Series = None,
                     feature_name: str = None,
                     report_type: ReportType = ReportType.SingleTimeSeries,
                     is_add_summary_table: bool = True,
                     file_name: str = 'econ_data',
                     var_format: Optional[str] = None,
                     is_price_data: bool = False,
                     desc_table_type: DescTableType = DescTableType.WITH_KURTOSIS,
                     x_min_max_quantiles: Tuple[Optional[float], Optional[float]] = (0.005, 0.995),
                     background: str = 'darkgrid',
                     **kwargs
                     ) -> List[plt.Figure]:
    """
    report for economic data
    """
    local_kwargs = {'fontsize': 6,
                    'linewidth': 0.5,
                    'weight': 'normal',
                    'framealpha': 0.75}
    qis.update_kwargs(local_kwargs, kwargs)

    figsize_table = (8.0, 2.9)  # table for 1 column
    figsize = (14, 10)

    if var_format is None:
        if is_price_data:
            var_format = '{:,.2f}'
        else:
            var_format = '{:.2%}'

    # get summary table
    price_start_dates = {}
    price_end_dates = {}
    insample_prices = {}
    for inst in data.columns:
        insample_price = dfo.drop_first_nan_data(df=data[inst]).dropna()

        if insample_price.empty:
            bbg_start = np.nan
            bbg_finish = np.nan
        else:
            bbg_start = insample_price.index[0]
            bbg_finish = insample_price.index[-1]

        price_start_dates.update({inst: bbg_start})
        price_end_dates.update({inst: bbg_finish})
        insample_prices.update({inst: insample_price})
    price_start_dates = pd.Series(price_start_dates).rename('StartDate')
    price_end_dates = pd.Series(price_end_dates).rename('EndDate')
    price_start_dates = pd.concat([price_start_dates, price_end_dates], axis=1)

    figs = []
    if is_add_summary_table:
        fig = get_summary_table_fig(data_start_dates=price_start_dates,
                                    descriptive_data=descriptive_data,
                                    figsize=figsize_table,
                                    **kwargs)
        figs.append(fig)

    group_means = None
    if group_data is not None:
        group_means = dfg.agg_df_by_groups(df=data,
                                           group_data=group_data,
                                           agg_func=dfa.nanmean,
                                           total_column=None)
    figure_num = 0
    for inst, insample_price in insample_prices.items():
        if np.all(insample_price.isnull()):  # skip for all nans
            continue
        figure_num += 1
        figure_caption = f"Figure {figure_num}. {inst}"
        if inst_names is not None:
            figure_caption += f", {inst_names[inst]}"

        if descriptive_data is not None:
            for k, v in descriptive_data.loc[inst].to_dict().items():
                figure_caption += f", {k}: {v}"

        local_kwargs = set_x_date_freq(data=insample_price, kwargs=kwargs)

        # change names
        if feature_name is None:
            inst_name = f"{inst} "
        else:
            insample_price = insample_price.rename(feature_name)
            inst_name = ""

        if complimentary_datas is not None:
            for name, data in complimentary_datas.items():
                if inst in data.columns:
                    comp_name = f"{inst_name}{name}"
                    insample_price = pd.concat([insample_price, data[inst].rename(comp_name)], axis=1)

        if group_means is not None:
            inst_group = group_data[inst]
            insample_price = pd.concat([insample_price, group_means[inst_group].rename(f"{inst_group} ac avg")], axis=1)

        if report_type == ReportType.SingleTimeSeries:
            with sns.axes_style(background):
                fig, ax = plt.subplots(1, 1, figsize=figsize, tight_layout=True)
                plot_data_timeseries(data=insample_price,
                                     ax=ax,
                                     title=f"In-sample period = ",
                                     is_price_data=is_price_data,
                                     var_format=var_format,
                                     **local_kwargs)

        elif report_type == ReportType.SingleTimeSeriesWithPDF:
            fig, axs = plt.subplots(1, 2, figsize=figsize, tight_layout=True)
            plot_data_timeseries(data=insample_price,
                                 ax=axs[0],
                                 title=f"In-sample period = ",
                                 is_price_data=is_price_data,
                                 var_format=var_format,
                                 **local_kwargs)

            pdf.plot_histogram(df=insample_price,
                               xvar_format=var_format,
                               x_min_max_quantiles=x_min_max_quantiles,
                               add_last_value=False,
                               desc_table_type=desc_table_type,
                               title=f"Distribution: {da.get_time_period(df=data).to_str()}",
                               bbox_to_anchor=None,
                               ax=axs[1],
                               **kwargs)

        else:
            raise TypeError(f"report_type={report_type} is not implemented")

        qis.set_suptitle(fig, title=f"{file_name}_{figure_num}_{inst}")
        figs.append(fig)

    return figs


class UnitTests(Enum):
    UPDATE_DATA = 1
    RUN_REPORT = 2


def run_unit_test(unit_test: UnitTests):

    import matplotlib.pyplot as plt
    import quant_strats.local_path as lp
    local_path = lp.get_resource_path()

    if unit_test == UnitTests.UPDATE_DATA:
        from bbg_fetch import fetch_field_timeseries_per_tickers
        vix_tickers = ['VIX1D Index', 'VIX9D Index', 'VIX Index', 'VIX3M Index', 'VIX6M Index', 'VIX1Y Index']
        vols = 0.01 * fetch_field_timeseries_per_tickers(tickers=vix_tickers, field='PX_LAST', CshAdjNormal=True)
        print(vols)
        qis.save_df_to_csv(df=vols, file_name='vix_indices', local_path=local_path)

    elif unit_test == UnitTests.RUN_REPORT:
        df = qis.load_df_from_csv(file_name='vix_indices', local_path=local_path)
        figs = econ_data_report(data=df)
        qis.save_figs_to_pdf(figs, file_name='vix_indices', local_path=local_path)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RUN_REPORT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
the systematic rolls for option or futures at a given timestamp include:
selection of the roll expiry given by parameter roll_freq (roll_freq='W-FRI', 'M-FRI')
roll time set using roll_hour
when the roll references second to first available roll expiry set by min_days_to_next_roll
example on how to generate generate roll schedule: for each timestamp find the corresponding maturity of the next roll
returned pd.DateTimeIndex is the current observation time and value is the corresponding option expiry of the current roll.
"""
import pandas as pd
pd.set_option('display.max_rows', 500)
import qis
from qis import TimePeriod

# set time period to fill with the roll dates
time_period = TimePeriod('01Jun2023', '08Sep2023', tz='UTC')

# weekly on Friday
weekly_rolls = qis.generate_fixed_maturity_rolls(time_period=time_period,
                                                 freq='D',  # frequency of the returned datetime index
                                                 roll_freq='W-FRI',  # roll expiries frequency
                                                 roll_hour=8,  # hour of the roll expiry
                                                 min_days_to_next_roll=6)  # days before maturity of the first contract before we switch to next contract
print(f"weekly_rolls:\n{weekly_rolls}")

# last friday of a month
monthly_rolls = qis.generate_fixed_maturity_rolls(time_period=time_period,
                                                  freq='D', roll_freq='M-FRI',
                                                  roll_hour=8, min_days_to_next_roll=28)  # 4 weeks before
print(f"monthly rolls:\n{monthly_rolls}")

# third friday of a month
friday3_monthly_rolls = qis.generate_fixed_maturity_rolls(time_period=time_period,
                                                          freq='D', roll_freq='WOM-3FRI',
                                                          roll_hour=8, min_days_to_next_roll=28)  # 4 weeks before expiry
print(f"monthly 3rd Friday rolls:\n{friday3_monthly_rolls}")

# third wednesday of a month
wednesday3_monthly_rolls = qis.generate_fixed_maturity_rolls(time_period=time_period,
                                                          freq='D', roll_freq='WOM-3WED',
                                                          roll_hour=8, min_days_to_next_roll=1)  # 4 weeks before expiry
print(f"monthly 3rd Wednesday rolls:\n{wednesday3_monthly_rolls}")


# last Friday of quarter
quarterly_rolls = qis.generate_fixed_maturity_rolls(time_period=time_period,
                                                    freq='D', roll_freq='Q-FRI',
                                                    roll_hour=8, min_days_to_next_roll=56)  # 8 weeks before expiry
print(f"quarterly rolls:\n{quarterly_rolls}")

# third Friday of quarter
quarterly_rolls = qis.generate_fixed_maturity_rolls(time_period=time_period,
                                                    freq='D', roll_freq='Q-3FRI',
                                                    roll_hour=8, min_days_to_next_roll=56)  # 8 weeks before expiry
print(f"quarterly 3rd Friday rolls:\n{quarterly_rolls}")



import pandas as pd
from typing import Tuple, List
from enum import Enum
import yfinance as yf
import qis
from qis import TimePeriod, PortfolioData
from qis.portfolio.reports.config import fetch_default_report_kwargs


def fetch_riskparity_universe_data() -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series]:
    """
    define custom universe with asset class grouping
    """
    universe_data = dict(SPY='Equities',
                         QQQ='Equities',
                         EEM='Equities',
                         TLT='Bonds',
                         IEF='Bonds',
                         LQD='Credit',
                         HYG='HighYield',
                         GLD='Gold')
    tickers = list(universe_data.keys())
    group_data = pd.Series(universe_data)  # for portfolio reporting
    prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Adj Close'][tickers]
    prices = prices.asfreq('B', method='ffill')  # make B frequency
    benchmark_prices = prices[['SPY', 'TLT']]
    return prices, benchmark_prices, group_data


def fetch_equity_bond() -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series]:
    """
    define custom universe with asset class grouping
    """
    universe_data = dict(SPY='Equities',
                         IEF='Bonds')
    tickers = list(universe_data.keys())
    group_data = pd.Series(universe_data)  # for portfolio reporting
    prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Adj Close'][tickers]
    benchmark_prices = prices[['SPY', 'IEF']]
    return prices, benchmark_prices, group_data


def generate_volparity_portfolio(prices: pd.DataFrame,
                                 group_data: pd.Series,
                                 time_period: TimePeriod = None,
                                 span: int = 60,
                                 vol_target: float = 0.15,
                                 rebalancing_costs: float = 0.0010
                                 ) -> PortfolioData:
    ra_returns, weights, ewm_vol = qis.compute_ra_returns(returns=qis.to_returns(prices=prices, is_log_returns=True),
                                                          span=span,
                                                          vol_target=vol_target)
    weights = weights.divide(weights.sum(1), axis=0)

    if time_period is not None:
        weights = time_period.locate(weights)

    volparity_portfolio = qis.backtest_model_portfolio(prices=prices,
                                                       weights=weights,
                                                       rebalancing_costs=rebalancing_costs,
                                                       is_output_portfolio_data=True,
                                                       ticker='VolParity')
    volparity_portfolio.set_group_data(group_data=group_data, group_order=list(group_data.unique()))
    return volparity_portfolio


def generate_equity_bond_portfolio(prices: pd.DataFrame,
                                   weights: List[float],
                                   group_data: pd.Series,
                                   rebalancing_costs: float = 0.0010
                                   ) -> PortfolioData:
    volparity_portfolio = qis.backtest_model_portfolio(prices=prices,
                                                       weights=weights,
                                                       rebalancing_costs=rebalancing_costs,
                                                       is_output_portfolio_data=True,
                                                       ticker='EquityBond')
    volparity_portfolio.set_group_data(group_data=group_data, group_order=list(group_data.unique()))
    return volparity_portfolio


class UnitTests(Enum):
    VOLPARITY_PORTFOLIO = 1
    EQUITY_BOND = 2


def run_unit_test(unit_test: UnitTests):

    time_period = qis.TimePeriod('31Dec2005', '31May2024')  # time period for portfolio reporting
    time_period_short = TimePeriod('31Dec2019', time_period.end)
    rebalancing_costs = 0.0010  # per traded volume

    if unit_test == UnitTests.VOLPARITY_PORTFOLIO:

        prices, benchmark_prices, group_data = fetch_riskparity_universe_data()
        portfolio_data = generate_volparity_portfolio(prices=prices,
                                                      group_data=group_data,
                                                      time_period=time_period,
                                                      span=30,
                                                      vol_target=0.15,
                                                      rebalancing_costs=rebalancing_costs)
        figs = qis.generate_strategy_factsheet(portfolio_data=portfolio_data,
                                               benchmark_prices=benchmark_prices,
                                               time_period=time_period,
                                               add_var_risk_sheet=True,
                                               **fetch_default_report_kwargs(time_period=time_period))
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"{portfolio_data.nav.name}_strategy_factsheet_long",
                             local_path=qis.local_path.get_output_path())
        """
        figs = qis.generate_strategy_factsheet(portfolio_data=portfolio_data,
                                               benchmark_prices=benchmark_prices,
                                               add_grouped_exposures=True,
                                               add_grouped_cum_pnl=True,
                                               time_period=time_period_short,
                                               **fetch_default_report_kwargs(time_period=time_period_short))
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"{portfolio_data.nav.name}_strategy_factsheet_short",
                             local_path=qis.local_path.get_output_path())
        
        qis.save_fig(fig=figs[0], file_name=f"strategy", local_path=qis.local_path.get_output_path())
        """

    elif unit_test == UnitTests.EQUITY_BOND:
        prices, benchmark_prices, group_data = fetch_equity_bond()
        portfolio_data = generate_equity_bond_portfolio(prices=prices,
                                                        weights=[0.6, 0.4],
                                                        group_data=group_data,
                                                        rebalancing_costs=rebalancing_costs)
        figs = qis.generate_strategy_factsheet(portfolio_data=portfolio_data,
                                               benchmark_prices=benchmark_prices,
                                               add_grouped_exposures=True,
                                               add_grouped_cum_pnl=True,
                                               time_period=time_period,
                                               **fetch_default_report_kwargs(time_period=time_period))
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"{portfolio_data.nav.name}_portfolio_factsheet_long",
                             local_path=qis.local_path.get_output_path())

        figs = qis.generate_strategy_factsheet(portfolio_data=portfolio_data,
                                               benchmark_prices=benchmark_prices,
                                               add_grouped_exposures=True,
                                               add_grouped_cum_pnl=True,
                                               time_period=TimePeriod('31Dec2019', time_period_short),
                                               **fetch_default_report_kwargs(time_period=time_period_short))
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"{portfolio_data.nav.name}_portfolio_factsheet_short",
                             local_path=qis.local_path.get_output_path())

    # plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VOLPARITY_PORTFOLIO

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
multi strategy backtest is generated using same strategy with a set of different model parameters
"""

import pandas as pd
import matplotlib.pyplot as plt
from typing import Tuple, List
from enum import Enum
import yfinance as yf
import qis
from qis import TimePeriod, MultiPortfolioData
from qis.portfolio.reports.multi_strategy_factsheet import generate_multi_portfolio_factsheet
from qis.portfolio.reports.config import fetch_default_report_kwargs


def fetch_riskparity_universe_data() -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series]:
    """
    define custom universe with asset class grouping
    """
    universe_data = dict(SPY='Equities',
                         QQQ='Equities',
                         EEM='Equities',
                         TLT='Bonds',
                         IEF='Bonds',
                         LQD='Credit',
                         HYG='HighYield',
                         GLD='Gold')
    tickers = list(universe_data.keys())
    group_data = pd.Series(universe_data)  # for portfolio reporting
    prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Adj Close'][tickers]
    prices = prices.asfreq('B', method='ffill').loc['2003': ]
    benchmark_prices = prices[['SPY', 'TLT']]
    return prices, benchmark_prices, group_data


def generate_volparity_multi_strategy(prices: pd.DataFrame,
                                      benchmark_prices: pd.DataFrame,
                                      group_data: pd.Series,
                                      time_period: TimePeriod,
                                      spans: List[int] = (5, 10, 20, 40, 60, 120),
                                      vol_target: float = 0.15,
                                      rebalancing_costs: float = 0.0010
                                      ) -> MultiPortfolioData:
    """
    generate volparity sensitivity to span
    """
    returns = qis.to_returns(prices=prices, is_log_returns=True)

    portfolio_datas = []
    for span in spans:
        ra_returns, weights, ewm_vol = qis.compute_ra_returns(returns=returns, span=span, vol_target=vol_target)
        weights = weights.divide(weights.sum(1), axis=0)
        portfolio_data = qis.backtest_model_portfolio(prices=prices,
                                                      weights=time_period.locate(weights),
                                                      rebalancing_costs=rebalancing_costs,
                                                      is_output_portfolio_data=True,
                                                      ticker=f"VP span={span}")
        portfolio_data.set_group_data(group_data=group_data, group_order=list(group_data.unique()))
        portfolio_datas.append(portfolio_data)

    multi_portfolio_data = MultiPortfolioData(portfolio_datas, benchmark_prices=benchmark_prices)
    return multi_portfolio_data


class UnitTests(Enum):
    VOLPARITY_SPAN = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.VOLPARITY_SPAN:

        time_period = qis.TimePeriod('31Dec2005', '31May2024')  # time period for portfolio reporting

        prices, benchmark_prices, group_data = fetch_riskparity_universe_data()
        multi_portfolio_data = generate_volparity_multi_strategy(prices=prices,
                                                                 benchmark_prices=benchmark_prices,
                                                                 group_data=group_data,
                                                                 time_period=time_period,
                                                                 vol_target=0.15,
                                                                 rebalancing_costs=0.0010  # per traded volume
                                                                 )

        figs = generate_multi_portfolio_factsheet(multi_portfolio_data=multi_portfolio_data,
                                                  time_period=time_period,
                                                  add_group_exposures_and_pnl=True,
                                                  **fetch_default_report_kwargs(time_period=time_period))

        qis.save_fig(fig=figs[0], file_name=f"multi_strategy", local_path=qis.local_path.get_output_path())

        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"volparity_span_factsheet_long",
                             orientation='landscape',
                             local_path=qis.local_path.get_output_path())

        time_period_short = TimePeriod('31Dec2019', time_period.end)
        figs = generate_multi_portfolio_factsheet(multi_portfolio_data=multi_portfolio_data,
                                                  time_period=time_period_short,
                                                  add_group_exposures_and_pnl=True,
                                                  **fetch_default_report_kwargs(time_period=time_period_short))
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"volparity_span_factsheet_short",
                             orientation='landscape',
                             local_path=qis.local_path.get_output_path())
    # plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VOLPARITY_SPAN

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
statsmodels wrappers for ols analysis
"""
# packages
import warnings
import numpy as np
import pandas as pd
from statsmodels import api as sm
from statsmodels.regression.linear_model import RegressionResults as RegModel
from typing import Tuple


def fit_ols(x: np.ndarray,
            y: np.ndarray,
            order: int = 1,
            fit_intercept: bool = True
            ) -> RegModel:
    """
    fit regression model
    """
    cond = np.logical_and(np.isfinite(x), np.isfinite(y))
    x, y = x[cond], y[cond]
    x1 = get_ols_x(x=x, order=order, fit_intercept=fit_intercept)
    reg_model = sm.OLS(y, x1).fit()
    return reg_model


def estimate_ols_alpha_beta(x: np.ndarray,
                            y: np.ndarray,
                            fit_intercept: bool = True
                            ) -> Tuple[float, float, float]:
    try:
        reg_model = fit_ols(x=x, y=y, order=1, fit_intercept=fit_intercept)
    except:
        warnings.warn(f"problem with x={x}, y={y}")
        return 0.0, 0.0, 0.0
    if fit_intercept:
        alpha = reg_model.params.iloc[0]
        beta = reg_model.params.iloc[1]
    else:
        alpha = 0.0
        beta = reg_model.params.iloc[0]
    r2 = reg_model.rsquared
    return alpha, beta, r2


def estimate_alpha_beta_paired_dfs(x: pd.DataFrame,
                                   y: pd.DataFrame,
                                   fit_intercept: bool = True
                                   ) -> Tuple[pd.Series, pd.Series]:
    """
    ols for paired x and y dfs, default axis=0
    """
    # align:
    x = x.dropna()
    y = y.dropna()
    y = y.loc[x.index, x.columns]
    x_np = x.to_numpy()
    y_np = y.to_numpy()
    n_col = len(x.columns)
    alphas, betas = np.zeros(n_col), np.zeros(n_col)
    for idx in np.arange(n_col):
        alphas[idx], betas[idx], _ = estimate_ols_alpha_beta(x=x_np[:, idx],
                                                             y=y_np[:, idx],
                                                             fit_intercept=fit_intercept)
    alphas = pd.Series(alphas, index=x.columns)
    betas = pd.Series(betas, index=x.columns)
    return alphas, betas


def get_ols_x(x: np.ndarray, order: int, fit_intercept: bool = True) -> np.ndarray:
    """
    compute powers of x
    """
    if order == 0:
        x = np.ones_like(x)
        fit_intercept = False
    elif order == 1:
        x = x
    elif order == 2:
        x = np.column_stack((x, np.square(x)))
    elif order == 3:
        x2 = np.square(x)
        x = np.column_stack((x, x2, x*x2))
    elif order == 4:
        x2 = np.square(x)
        x = np.column_stack((x, x2, x*x2, x2*x2))
    else:
        raise ValueError(f"order = {order} is not implemnted")

    if fit_intercept:
        x = sm.add_constant(x)
    return x


def reg_model_params_to_str(reg_model: RegModel,
                            order: int,
                            r2_only: bool = False,
                            beta_format: str = '{0:+0.2f}',
                            alpha_format: str = '{0:+0.2f}',
                            fit_intercept: bool = True,
                            alpha_an_factor: float = None,
                            **kwargs
                            ) -> str:
    try:
        r2 = f", R\N{SUPERSCRIPT TWO}={reg_model.rsquared:.0%}"
    except:
        r2 = f", R\N{SUPERSCRIPT TWO}=0.0%"

    if r2_only:
        text_str = f" R\N{SUPERSCRIPT TWO}={reg_model.rsquared:.0%}"
    else:
        if fit_intercept:
            if alpha_an_factor is not None:
                # alpha = '{:+0.0%}'.format(alpha_an_factor*reg_model.params[0])
                alpha = '{:+0.0%}'.format(np.expm1(alpha_an_factor * reg_model.params[0]))
            else:
                alpha = alpha_format.format(reg_model.params[0])
            idx1 = 1
        else:
            alpha = ''
            idx1 = 0

        if order == 1:
            text_str = 'y=' + beta_format.format(reg_model.params[idx1]) + 'X' \
                       + alpha \
                       + f', R\N{SUPERSCRIPT TWO}=' + '{0:.0%}'.format(reg_model.rsquared)

        elif order == 2:
            if fit_intercept:  # with intercept
                text_str = 'y=' + beta_format.format(reg_model.params[idx1+1]) + 'X' + f'\N{SUPERSCRIPT TWO}' \
                            + beta_format.format(reg_model.params[idx1]) + 'X' \
                            + alpha + r2
            else :  # without intercept
                text_str = 'y=' + beta_format.format(reg_model.params[idx1+1]) + 'X' + f'\N{SUPERSCRIPT TWO}' \
                            + beta_format.format(reg_model.params[idx1]) + 'X' \
                            + alpha \
                            + f', R\N{SUPERSCRIPT TWO}=' + '{0:.0%}'.format(reg_model.rsquared)

        elif order == 3:
            try:
                text_str = 'y=' + beta_format.format(reg_model.params[idx1+2]) + 'x' + f'\N{SUPERSCRIPT THREE}' \
                           + beta_format.format(reg_model.params[idx1+1]) + 'x' + f'\N{SUPERSCRIPT TWO}' \
                           + beta_format.format(reg_model.params[idx1]) + 'x' \
                           + alpha \
                           + f', R\N{SUPERSCRIPT TWO}=' + '{0:.0%}'.format(reg_model.rsquared)
            except:
                text_str = 'model cannot be estimated'
        else:
            raise TypeError(f"order = {order} is not implemnted")

    return text_str



"""
performance report for a universe of several assets
with comparison to 1-2 benchmarks
output is one-page figure with key numbers
"""
# packages
import matplotlib.pyplot as plt
import yfinance as yf
import qis as qis
from enum import Enum
from qis.portfolio.reports.multi_assets_factsheet import generate_multi_asset_factsheet
from qis.portfolio.reports.config import fetch_default_report_kwargs


class UnitTests(Enum):
    CORE_ETFS = 1
    BTC_SQQQ = 2
    HEDGED_ETFS = 3
    BBG = 4
    RATES_FUTURES = 5
    HYG_ETFS = 6


def run_unit_test(unit_test: UnitTests):

    prices = None

    if unit_test == UnitTests.CORE_ETFS:
        benchmark = 'SPY'
        tickers = [benchmark, 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD']
        time_period = qis.TimePeriod('31Dec2007', '31May2024')  # time period for reporting

    elif unit_test == UnitTests.BTC_SQQQ:
        benchmark = 'QQQ'
        tickers = [benchmark, 'BTC-USD', 'TQQQ', 'SQQQ']
        time_period = qis.TimePeriod('31Dec2019', '31May2024')

    elif unit_test == UnitTests.HEDGED_ETFS:
        benchmark = 'SPY'
        tickers = [benchmark, 'SHY', 'LQDH', 'HYGH', 'FLOT']
        time_period = qis.TimePeriod('27May2014', '31May2024')

    elif unit_test == UnitTests.BBG:
        benchmark = 'SPTR'
        tickers = {'SPTR Index': benchmark,
                   'SGEPMLU Index': 'SG AI long/short', 'SGEPMLUL Index': 'SG AI long', 'SGEPMLUS Index': 'SG AI short',
                   'SGIXAI Index': 'SG Alpha',
                   'AIPEX Index': 'HSBC Aipex',
                   'MQQFUSAI Index': 'MerQube AI',
                   'CSRPAIS Index': 'CS RavenPack',
                   'CITIMSTR Index': 'Citi Grand'}
        time_period = qis.TimePeriod('31Dec2019', '20Jun2024')
        from bbg_fetch import fetch_field_timeseries_per_tickers
        prices = fetch_field_timeseries_per_tickers(tickers=list(tickers.keys()), field='PX_LAST', CshAdjNormal=True).dropna()
        prices = prices.rename(tickers, axis=1)

    elif unit_test == UnitTests.RATES_FUTURES:
        benchmark = '2y UST'
        tickers = {'TU1 Comdty': benchmark,
                   'ED5 Comdty': 'USD IR',
                   'L 5 Comdty': 'GBP IR',
                   'ER4 Comdty': 'EUR IR',
                   'IR4 Comdty': 'AUD IR',
                   'BA4 Comdty': 'CAD IR'}
        time_period = qis.TimePeriod(start='02Apr1986', end='31Mar2023')
        from bbg_fetch import fetch_field_timeseries_per_tickers
        prices = fetch_field_timeseries_per_tickers(tickers=list(tickers.keys()), field='PX_LAST', CshAdjNormal=True).dropna()
        prices = prices.rename(tickers, axis=1)

    elif unit_test == UnitTests.HYG_ETFS:
        benchmark = 'IBOXHY'
        tickers = {'IBOXHY Index': benchmark,
                   'HYDB US Equity': 'HYDB',
                   'HYG US Equity': 'HYG'
                   }
        from bbg_fetch import fetch_field_timeseries_per_tickers
        prices = fetch_field_timeseries_per_tickers(tickers=list(tickers.keys()), field='PX_LAST', CshAdjNormal=True).dropna()
        prices = prices.rename(tickers, axis=1)
        time_period = qis.get_time_period(prices)


    else:
        raise NotImplementedError

    if prices is None:
        prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Adj Close'][tickers]

    prices = prices.asfreq('B', method='ffill')  # make B frequency
    prices = prices.dropna()
    print(prices)

    kwargs = fetch_default_report_kwargs(time_period=time_period)

    fig = generate_multi_asset_factsheet(prices=prices,
                                         benchmark=benchmark,
                                         time_period=time_period,
                                         **kwargs)
    qis.save_figs_to_pdf(figs=[fig],
                         file_name=f"multiasset_report", orientation='landscape',
                         local_path=qis.local_path.get_output_path())
    qis.save_fig(fig=fig, file_name=f"multiassets", local_path=qis.local_path.get_output_path())

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.HYG_ETFS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
example of using startegy construction and reporting
for implementation see portfolio.report.strategy_factsheet
"""
# packages
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple
from enum import Enum
import yfinance as yf
import qis as qis
from qis import TimePeriod, MultiPortfolioData

# reporting
from qis.portfolio.reports.config import fetch_default_report_kwargs
from qis.portfolio.reports.strategy_benchmark_factsheet import (generate_strategy_benchmark_factsheet_plt,
                                                                generate_strategy_benchmark_active_perf_plt,
                                                                generate_performance_attribution_report)


def fetch_riskparity_universe_data() -> Tuple[pd.DataFrame, pd.DataFrame, pd.Series]:
    """
    define custom universe with asset class grouping
    """
    universe_data = dict(SPY='Equities',
                         QQQ='Equities',
                         EEM='Equities',
                         TLT='Bonds',
                         IEF='Bonds',
                         LQD='Credit',
                         HYG='HighYield',
                         GLD='Gold')
    tickers = list(universe_data.keys())
    group_data = pd.Series(universe_data)  # for portfolio reporting
    prices = yf.download(tickers=tickers, start=None, end=None, ignore_tz=True)['Adj Close'][tickers]
    prices = prices.asfreq('B', method='ffill')
    benchmark_prices = prices[['SPY', 'TLT']]
    return prices, benchmark_prices, group_data


def generate_volparity_multiportfolio(prices: pd.DataFrame,
                                      benchmark_prices: pd.DataFrame,
                                      group_data: pd.Series,
                                      time_period: TimePeriod = None,
                                      span: int = 60,
                                      vol_target: float = 0.15,
                                      rebalancing_costs: float = 0.0010
                                      ) -> MultiPortfolioData:

    ra_returns, weights, ewm_vol = qis.compute_ra_returns(returns=qis.to_returns(prices=prices, is_log_returns=True),
                                                          span=span,
                                                          vol_target=vol_target)
    weights = weights.divide(weights.sum(1), axis=0)

    if time_period is not None:
        weights = time_period.locate(weights)

    volparity_portfolio = qis.backtest_model_portfolio(prices=prices,
                                                       weights=weights,
                                                       rebalancing_costs=rebalancing_costs,
                                                       is_output_portfolio_data=True,
                                                       ticker='VolParity')
    volparity_portfolio.set_group_data(group_data=group_data, group_order=list(group_data.unique()))

    ew_portfolio = qis.backtest_model_portfolio(prices=prices,
                                                weights=np.ones(len(prices.columns)) / len(prices.columns),
                                                rebalancing_costs=rebalancing_costs,
                                                is_output_portfolio_data=True,
                                                ticker='EqualWeight')
    ew_portfolio.set_group_data(group_data=group_data, group_order=list(group_data.unique()))

    multi_portfolio_data = MultiPortfolioData(portfolio_datas=[volparity_portfolio, ew_portfolio],
                                              benchmark_prices=benchmark_prices)
    return multi_portfolio_data


class UnitTests(Enum):
    STRATEGY_BENCHMARK_PLT = 1
    PERFORMANCE_ATTRIBUTION = 2
    ACTIVE_PERFORMANCE = 3


def run_unit_test(unit_test: UnitTests):

    # time period for portfolio reporting
    time_period = qis.TimePeriod('31Dec2006', '31May2024')
    prices, benchmark_prices, group_data = fetch_riskparity_universe_data()
    multi_portfolio_data = generate_volparity_multiportfolio(prices=prices,
                                                             benchmark_prices=benchmark_prices,
                                                             group_data=group_data,
                                                             time_period=time_period,
                                                             span=30,
                                                             vol_target=0.15,
                                                             rebalancing_costs=0.0010  # per traded volume
                                                             )

    if unit_test == UnitTests.STRATEGY_BENCHMARK_PLT:
        pnl_attribution = True
        figs = generate_strategy_benchmark_factsheet_plt(multi_portfolio_data=multi_portfolio_data,
                                                         backtest_name='Vol Parity Portfolio vs Equal Weight',
                                                         time_period=time_period,
                                                         add_brinson_attribution=pnl_attribution,
                                                         add_exposures_pnl_attribution=pnl_attribution,
                                                         add_strategy_factsheet=False, # for strategy factsheet
                                                         add_grouped_exposures=False,  # for strategy factsheet
                                                         add_grouped_cum_pnl=False,  # for strategy factsheet
                                                         **fetch_default_report_kwargs(time_period=time_period))
        """

        """
        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"strategy_benchmark_factsheet", orientation='landscape',
                             local_path=qis.local_path.get_output_path())
        qis.save_fig(fig=figs[0], file_name=f"strategy_benchmark", local_path=qis.local_path.get_output_path())
        if pnl_attribution:
            qis.save_fig(fig=figs[1], file_name=f"brinson_attribution", local_path=qis.local_path.get_output_path())
            qis.save_fig(fig=figs[2], file_name=f"pnl_attribution", local_path=qis.local_path.get_output_path())

    elif unit_test == UnitTests.PERFORMANCE_ATTRIBUTION:
        figs = generate_performance_attribution_report(multi_portfolio_data=multi_portfolio_data,
                                                       time_period=TimePeriod('31Dec2021', None),
                                                       **fetch_default_report_kwargs(time_period=time_period))

    elif unit_test == UnitTests.ACTIVE_PERFORMANCE:
        figs = generate_strategy_benchmark_active_perf_plt(multi_portfolio_data=multi_portfolio_data,
                                                           time_period=time_period,
                                                           figsize=(11, 6),
                                                           is_long_only=True,
                                                           **fetch_default_report_kwargs(time_period=time_period))

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.STRATEGY_BENCHMARK_PLT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
import qis.plots.derived as put
"""

# packages
import colorsys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import seaborn as sns
from matplotlib._color_data import CSS4_COLORS as mcolors
from matplotlib.colors import rgb2hex, LinearSegmentedColormap
from matplotlib.container import BarContainer
from matplotlib.lines import Line2D
from scipy import stats as stats
from scipy.stats import skew, kurtosis
from enum import Enum
from typing import List, Union, Tuple, Optional, Dict, Any

# qis
import qis.utils.df_ops as dfo
import qis.utils.df_str as dfs
import qis.utils.df_freq as dff
import qis.utils.struct_ops as sop
from qis.plots.table import ROW_HIGHT, COLUMN_WIDTH, FIRST_COLUMN_WIDTH


class FixedColors(Enum):
    NAVY = mcolors['navy']
    DARKGREEN = mcolors['darkgreen']
    DARKRED = mcolors['darkred']
    CHOCOLATE = mcolors['chocolate']
    DODGERBLUE = mcolors['slateblue']
    DIMGRAY = mcolors['dimgray']
    SANDYBROWN = mcolors['deepskyblue']
    SADDLEBROWN = mcolors['saddlebrown']
    PURPLE = mcolors['purple']
    INDIANRED = mcolors['indianred']
    FIREBRICK = mcolors['firebrick']
    DARKORANGE = mcolors['darkorange']
    MAGENTA = mcolors['magenta']
    PEACHPUFF = mcolors['deepskyblue']
    SIENNA = mcolors['sienna']
    PERU = mcolors['peru']
    LIGHTCORAL = mcolors['lightcoral']
    CRIMSON = mcolors['crimson']


class TrendLine(Enum):
    NONE = 0
    AVERAGE = 1
    AVERAGE_SHADOWS = 2
    ZERO = 3
    ZERO_SHADOWS = 4
    TREND_LINE = 5
    TREND_LINE_SHADOWS = 6
    ABOVE_ZERO_SHADOWS = 7


class LastLabel(Enum):
    NONE = 0
    LAST_VALUE = 1
    LAST_VALUE_SORTED = 2
    AVERAGE_VALUE = 3
    AVERAGE_VALUE_SORTED = 4


def create_dummy_line(**kwargs) -> Line2D:
    return Line2D([], [], **kwargs)


def set_title(ax: plt.Subplot,
              title: str,
              fontsize: int = 12,
              title_color: str = 'darkblue',
              pad: int = None,
              **kwargs
              ) -> None:
    ax.set_title(label=title, color=title_color, fontsize=fontsize, pad=pad)


def set_suptitle(fig: plt.Figure,
                 title: str,
                 fontsize: int = 12,
                 color: str = 'dodgerblue',
                 **kwargs
                 ) -> None:
    fig.suptitle(title, color=color, fontsize=fontsize)


def set_ax_tick_params(ax: plt.Subplot,
                       fontsize: int = None,
                       labelbottom: bool = True,
                       labelleft: bool = True,
                       labeltop: bool = False,
                       pad: int = 0,
                       **kwargs
                       ) -> None:
    ax.tick_params(axis='both',
                   which='both',
                   bottom=False, left=False, top=False, right=False,  # disable ticks
                   labelbottom=labelbottom,
                   labelleft=labelleft,
                   labeltop=labeltop,
                   labelsize=fontsize,
                   pad=pad)


def set_ax_ticks_format(ax: plt.Subplot,
                        xvar_format: Optional[str] = None,
                        yvar_format: str = None,
                        fontsize: int = 10,
                        set_ticks: bool = True,
                        xvar_numticks: int = None,
                        xvar_major_ticks: np.ndarray = None,
                        yvar_numticks: int = None,
                        yvar_major_ticks: np.ndarray = None,
                        x_rotation: int = 0,
                        **kwargs
                        ) -> None:
    if set_ticks:
        set_ax_tick_params(ax=ax, **kwargs)  # always remove minor ticks
    # set x - ticks
    ax.xaxis.set_minor_formatter(mticker.NullFormatter())  # always remove minor ticks

    # Tick locators
    if xvar_numticks is not None:
        ax.xaxis.set_major_locator(mticker.LinearLocator(numticks=xvar_numticks))
    elif xvar_major_ticks is not None:
        ax.xaxis.set_major_locator(mticker.FixedLocator(xvar_major_ticks))
    else:
        ticks_loc = ax.get_xticks()
        ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))

    if yvar_numticks is not None:
        ax.yaxis.set_major_locator(mticker.LinearLocator(numticks=yvar_numticks))
    elif yvar_major_ticks is not None:
        ax.yaxis.set_major_locator(mticker.FixedLocator(yvar_major_ticks))

    # Tick formatters
    if xvar_format is not None:
        ax.xaxis.set_major_formatter(mticker.FuncFormatter(lambda z, _: xvar_format.format(z)))

    for z in ax.get_xticklabels():
        z.set_fontsize(fontsize=fontsize)

    # set y - ticks
    ax.yaxis.set_minor_formatter(mticker.NullFormatter())  # always remove minor ticks
    if yvar_format is not None:
        ax.yaxis.set_major_formatter(mticker.FuncFormatter(lambda z, _: yvar_format.format(z)))
    for z in ax.get_yticklabels():
        z.set_fontsize(fontsize=fontsize)
    ax.tick_params(axis='x', rotation=x_rotation)


def set_ax_xy_labels(ax: plt.Subplot,
                     fontsize: int = 10,
                     xlabel: str = None,
                     ylabel: str = None,
                     labelpad: float = None,
                     **kwargs
                     ) -> None:

    if xlabel is not None:
        ax.set_xlabel(xlabel, fontsize=fontsize)
    else:
        ax.set_xlabel('')

    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=fontsize, labelpad=labelpad)
    else:
        ax.set_ylabel('')


def set_ax_tick_labels(ax: plt.Subplot,
                       x_labels: list = None,
                       y_labels: list = None,
                       x_rotation: int = 90,
                       y_rotation: int = 0,
                       is_remove_labels: bool = False,
                       xticks: List[float] = None,
                       fontsize: int = 10,
                       skip_y_axis: bool = False,
                       **kwargs
                       ) -> None:

    if is_remove_labels:
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    else:

        if xticks is not None:
            ax.set_xticks(xticks)
        else:  # to stop matplotlib 3.3.1 warning
            ax.set_xticks(ax.get_xticks())

        if x_labels is None:
            x_labels = ax.get_xticklabels()

        ax.set_xticklabels(x_labels,
                           fontsize=fontsize,
                           rotation=x_rotation,
                           minor=False)

        if not skip_y_axis:
            # to stop matplotlib 3.3.1 warning
            ax.set_yticks(ax.get_yticks())

            if y_labels is None:
                y_labels = ax.get_yticklabels()
            ax.set_yticklabels(y_labels,
                               fontsize=fontsize,
                               rotation=y_rotation,
                               minor=False)
        # else:
        #    ax.yaxis.set_major_formatter(plt.NullFormatter())
        #    ax.set_yticklabels([])


def validate_returns_plot(prices: Union[pd.DataFrame, pd.Series],
                          min_number: int = 20,  # 4 years of q returns
                          freq: str = 'QE',
                          fontsize: int = 8,
                          ax: plt.Figure = None,
                          **kwargs
                          ) -> Tuple[bool, Optional[plt.Figure]]:
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 4))
    else:
        fig = None

    # num_returns = len(prices.asfreq(freq, method='ffill').dropna().index)
    num_returns = len(prices.asfreq(freq, method='ffill').index)
    if num_returns < min_number:
        is_good = False
        text = f"the number of available returns = {str(num_returns)}\n is not sufficient to make figure"
        ax.text(0.1, 0.5, text,
                transform=ax.transAxes,
                style='italic',
                fontsize=fontsize,
                bbox={'facecolor': 'red', 'alpha': 0.5, 'pad': 10})
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
    else:
        is_good = True

    return is_good, fig


def set_spines(ax: plt.Subplot,
               top_spine: bool = False,
               bottom_spine: bool = True,
               left_spine: bool = True,
               right_spine: bool = False,
               **kwargs
               ) -> None:
    ax.spines['top'].set_visible(top_spine)
    ax.spines['bottom'].set_visible(bottom_spine)
    ax.spines['left'].set_visible(left_spine)
    ax.spines['right'].set_visible(right_spine)


def set_ax_linewidth(ax: plt.Subplot, linewidth: float = 2.0):
    for line in ax.get_lines():
        line.set_linewidth(linewidth)
    for line in ax.get_legend().get_lines():
        line.set_linewidth(linewidth)


def remove_spines(ax: plt.Subplot) -> None:
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)


def subplot_border(fig: plt.Figure,
                   nrows: int = 1,
                   ncols: int = 1,
                   color: str = 'navy'
                   ) -> None:
    """
    draw border for a figure with multiple subplots
    """

    n_ax1 = nrows
    n_ax2 = ncols

    rects = []
    height = 1.0 / n_ax1
    for r in range(n_ax1):
        rects.append(plt.Rectangle((0.0, r*height), 1.0, height,  # (lower-left corner), width, height
                                   fill=False,
                                   color=color,
                                   lw=1,
                                   zorder=1000,
                                   transform=fig.transFigure, figure=fig))
    width = 1.0 / n_ax2
    for r in range(n_ax2):
        rects.append(plt.Rectangle((r*width, 0), width, 1.0,  # (lower-left corner), width, height
                                   fill=False,
                                   color=color,
                                   lw=1,
                                   zorder=1000,
                                   transform=fig.transFigure, figure=fig))
    fig.patches.extend(rects)


def autolabel(ax: plt.Subplot,
              rects: List[BarContainer],
              xpos: str = 'center',
              name: str = '',
              label0: str = None,
              color: str = None,
              shapecolor: str = None,
              fontsize: int = 10,
              y_offset: int = 8,
              **kwargs
              ) -> None:
    """
    Attach a text label above each bar in *rects*, displaying its height.
    *xpos* indicates which side to place the text w.r.t. the center of
    the bar. It can be one of the following {'center', 'right', 'left'}.
    """

    ha = {'center': 'center', 'right': 'left', 'left': 'right'}
    offset = {'center': 0, 'right': 1, 'left': -1}

    for rect in rects:
        height = rect.get_height()
        y_offset = y_offset if height > 0.0 else -2.0*y_offset

        if shapecolor is not None:
            bbox = dict(boxstyle='round,pad=0.5', fc=shapecolor, alpha=0.5)
        else:
            bbox = None

        if label0 is None:
            label = f"{name} {'{0:.1%}'.format(height)}"
        else:
            label = label0

        ax.annotate(label,
                    xy=(rect.get_x() + rect.get_width() / 2, height),
                    xytext=(offset[xpos], y_offset),
                    textcoords="offset points",  # in both directions
                    fontsize=fontsize,
                    color=color,
                    ha=ha[xpos], va='bottom',
                    bbox=bbox)


def set_legend_colors(ax: plt.Subplot,
                      text_weight: str = None,
                      colors: List[str] = None,
                      fontsize: int = 12,
                      **kwargs
                      ) -> None:

    leg = ax.get_legend()
    if colors is None:
        colors = [line.get_color() for line in leg.get_lines()]

    for text, color in zip(leg.get_texts(), colors):
        text.set_color(color)
        text.set_size(fontsize)
        if text_weight is not None:
            text.set_weight(text_weight)


def set_y_limits(ax: plt.Subplot,
                 y_limits: Tuple[Optional[float], Optional[float]]
                 ) -> None:

    ymin, ymax = ax.get_ylim()
    if y_limits[0] is not None:
        ymin = y_limits[0]
    if y_limits[1] is not None:
        ymax = y_limits[1]
    ax.set_ylim([ymin, ymax])


def set_x_limits(ax: plt.Subplot,
                 x_limits: Tuple[Optional[float], Optional[float]]
                 ) -> None:
    xmin, xmax = ax.get_xlim()
    if x_limits[0] is not None:
        xmin = x_limits[0]
    if x_limits[1] is not None:
        xmax = x_limits[1]
    ax.set_xlim([xmin, xmax])


def align_xy_limits(ax: plt.Subplot) -> None:
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    xy_min = np.minimum(xmin, ymin)
    xy_max = np.maximum(xmax, ymax)
    ax.set_xlim([xy_min, xy_max])
    ax.set_ylim([xy_min, xy_max])


def align_y_limits_ax12(ax1: plt.Subplot,
                        ax2: plt.Subplot,
                        is_invisible_x_ax1: bool = False,
                        is_invisible_y_ax2: bool = False,
                        ymin: float = None,
                        ymax: float = None,
                        **kwargs
                        ) -> None:
    ymin1, ymax1 = ax1.get_ylim()
    ymin2, ymax2 = ax2.get_ylim()
    if ymin is None:
        ymin = np.minimum(ymin1, ymin2)
    if ymax is None:
        ymax = np.maximum(ymax1, ymax2)

    ax1.set_ylim([ymin, ymax])
    ax2.set_ylim([ymin, ymax])

    if is_invisible_x_ax1:
        ax1.axes.get_xaxis().set_visible(False)
    if is_invisible_y_ax2:
        ax2.axes.get_yaxis().set_visible(False)


def align_y_limits_axs(axs: List[plt.Subplot],
                       is_invisible_ys: bool = False
                       ) -> None:
    ymins = []
    ymaxs = []
    for ax in axs:
        ymin, ymax = ax.get_ylim()
        ymins.append(ymin)
        ymaxs.append(ymax)
    ymin = np.min(ymins)
    ymax = np.max(ymaxs)
    for ax in axs:
        ax.set_ylim([ymin, ymax])

    if is_invisible_ys:
        for idx, ax in enumerate(axs):
            if idx > 0:
                ax.axes.get_yaxis().set_visible(False)


def align_x_limits_ax12(ax1: plt.Subplot,
                        ax2: plt.Subplot,
                        is_invisible_x_ax1: bool = False,
                        is_invisible_y_ax2: bool = False,
                        **kwargs
                        ) -> None:

    xmin1, xmax1 = ax1.get_xlim()
    xmin2, xmax2 = ax2.get_xlim()
    xmin = np.minimum(xmin1, xmin2)
    xmax = np.maximum(xmax1, xmax2)
    ax1.set_xlim([xmin, xmax])
    ax2.set_xlim([xmin, xmax])

    if is_invisible_x_ax1:
        ax1.axes.get_xaxis().set_visible(False)
    if is_invisible_y_ax2:
        ax2.axes.get_yaxis().set_visible(False)


def align_x_limits_axs(axs: List[plt.Subplot],
                       is_invisible_xs: bool = False
                       ) -> None:
    xmins = []
    xmaxs = []
    for ax in axs:
        xmin, xmax = ax.get_xlim()
        xmins.append(xmin)
        xmaxs.append(xmax)
    xmin = np.min(xmins)
    xmax = np.max(xmaxs)
    for ax in axs:
        ax.set_xlim([xmin, xmax])

    if is_invisible_xs:
        for idx, ax in enumerate(axs):
            if idx < len(axs)-1:
                ax.axes.get_xaxis().set_visible(False)


def set_date_on_axis(data: Union[pd.DataFrame, pd.Series],
                     ax: plt.Subplot,
                     x_date_freq: str = None,
                     is_set_date_minmax: bool = False,
                     date_format: str = '%b-%y',
                     x_date_rotation: int = 90,
                     fontsize: int = 10,
                     **kwargs
                     ) -> None:

    if is_set_date_minmax and isinstance(data.index, pd.DatetimeIndex):
        datemin = np.datetime64(data.index[0], 'Y')
        datemax = np.datetime64(data.index[-1], 'Y') + np.timedelta64(1, 'Y')
        ax.set_xlim(datemin, datemax)

    if isinstance(data.index, pd.DatetimeIndex):
        if x_date_freq is None:
            labels = data.index.strftime(date_format).to_list()  # specified by index converted to strings
        else:
            ticks = pd.date_range(start=data.index[0], end=data.index[-1], freq=x_date_freq)
            if len(ticks) == 1:
                # try to find optimal freq:
                choices = ['ME', 'W-WED', 'B']
                for choice in choices:
                    ticks = pd.date_range(start=data.index[0], end=data.index[-1], freq=choice)
                    if len(ticks) > 1:
                        break
            labels = ticks.strftime(date_format).to_list()
            ax.set_xticks(ticks)

    else:
        labels = data.index
        x_date_rotation = 0
        ax.set_xticks(labels)

    # remove ticks
    # set_ax_ticks(ax=ax)
    ticks_loc = ax.get_xticks().tolist()
    ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    ax.set_xticklabels(labels,
                       fontsize=fontsize,
                       rotation=x_date_rotation,
                       weight='normal',
                       minor=False)
    # remove minor tick labels put by matplotlib
    ax.xaxis.set_minor_formatter(mticker.NullFormatter())


def map_dates_index_to_str(data: Union[pd.DataFrame, pd.Series],
                           x_date_freq: str = 'YE',
                           date_format: str = '%b-%y'
                           ) -> Tuple[pd.DataFrame, List[str]]:
    """
    get labels for barplot: needed when index is datetime and we want to display in dateformat
    """
    re_indexed_data = data.copy()
    datalables = data.index

    if x_date_freq is None:
        # x_date_freq = None corresponds to string labelled data
        pass
    elif data.index.dtype == 'str':
        # no map is needed for string index
        pass
    elif isinstance(data.index, pd.DatetimeIndex):
        # map dates to to str
        dates_index = list(pd.to_datetime(data.index))

        ticks = pd.date_range(start=data.index[0], end=data.index[-1], freq=x_date_freq)
        if len(ticks) == 1:
            # try to find optimal freq for less than annual
            choices = ['ME', 'W-WED', 'B']
            for choice in choices:
                ticks = pd.date_range(start=data.index[0], end=data.index[-1], freq=choice)
                if len(ticks) > 1:
                    x_date_freq = choice
                    break

        if x_date_freq == 'YE':
            re_indexed_data.index = [t.strftime(date_format) for t in dates_index]
            if pd.infer_freq(data.index) is not None:
                datalables = [t.strftime(date_format) if t.month == 12 else '' for t in dates_index]
            else:  # uneven dates so get 1y apart
                datalables = [t0.strftime(date_format) if t.year - t0.year == 1 and t.year % 2 == 0 else '' for
                              t, t0 in zip(dates_index[1:], dates_index[:-1])] + [dates_index[-1].strftime(date_format)]
        elif x_date_freq == 'A-Mar':
            re_indexed_data.index = [t.strftime(date_format) for t in dates_index]
            datalables = [t.strftime(date_format) if t.month == 3 else '' for t in dates_index]  # and t.year % 2 == 0

        elif x_date_freq == 'QE':
            re_indexed_data.index = [t.strftime(date_format) for t in dates_index]
            if pd.infer_freq(data.index) is not None:
                datalables = [t.strftime(date_format) if t.month % 3 == 0 else '' for t in dates_index]
            else:  # uneven dates so get 1y apart
                datalables = [t0.strftime(date_format) if t.month - t0.month == 1 and t0.month % 3 == 0 else '' for
                              t, t0 in zip(dates_index[1:], dates_index[:-1])] + [dates_index[-1].strftime(date_format)]

        elif x_date_freq == 'QS':
            indices = pd.Series(pd.to_datetime(data.index), index=data.index)
            indices = dff.df_asfreq(indices, freq=x_date_freq, inclusive='right', include_end_date=False).to_list()
            re_indexed_data.index = [t.strftime(date_format) for t in dates_index]
            datalables = [t.strftime(date_format) if t in indices else '' for t in dates_index]

        elif x_date_freq == 'B':
            re_indexed_data.index = [t.strftime('%b-%d') for t in dates_index]
            datalables = [t0.strftime('%b-%d') if t.day - t0.day > 0 else '' for
             t, t0 in zip(dates_index[1:], dates_index[:-1])] + [dates_index[-1].strftime('%b-%d')]

        else:  # does not matter
            datalables = [t.strftime('%d-%b-%y') for t in dates_index]
            re_indexed_data.index = datalables

            if x_date_freq == 'ME':
                datalables = [t0.strftime(date_format) if t.month - t0.month == 1 or t.month - t0.month == -11 else '' for
                              t, t0 in zip(dates_index[1:], dates_index[:-1])] + [dates_index[-1].strftime(date_format)]

            elif x_date_freq == '5A':
                datalables = [t.strftime(date_format) if t.month == 12 and t.year % 5 == 0 else '' for t in dates_index]

            elif list(x_date_freq)[-1] == 'YE': # frequncy of type 1A, 2A, 3A,...
                n_years = int(sop.separate_number_from_string(x_date_freq)[0])
                datalables = [t.strftime(date_format) if t.month == 12 and t.year % n_years == 0 else '' for t in dates_index]

        # remove dublicates
        val0 = ''
        for ind, val in enumerate(datalables):
            if val == val0:
                datalables[ind] = ''
            val0 = val

    return re_indexed_data, datalables


def set_legend(ax: plt.Subplot,
               labels: Union[List[str], pd.Index] = None,
               lines: List[Tuple] = None,
               markers: List[str] = None,
               colors: Optional[List[str]] = None,
               legend_loc: str = 'upper left',
               reverse_columns: bool = False,
               bbox_to_anchor: Tuple[float, float] = None,
               text_weight: str = 'light',
               legend_title: Optional[str] = None,
               fontsize: int = 12,
               ncol: int = 1,
               framealpha: float = 0.0,
               handlelength: float = 1.0,
               facecolor: str = None,
               numpoints: int = None,
               **kwargs
               ) -> None:

    if legend_loc is None:
        ax.legend([]).set_visible(False)
        return

    if labels is None:  # this are tuples of 3
        leg = ax.get_legend()
        if leg is not None:
            labels = [label.get_text() for label in leg.get_texts()]
        elif lines is None:
            print('in set_legend: cannot put labels from empty line')
            return

    if colors is None:
        leg = ax.get_legend()
        if leg is not None:
            colors = [line.get_color() for line in leg.get_lines()]

    if lines is None:
        lines = []
        is_markers_list = False
        if markers is not None and isinstance(markers, list):
            is_markers_list = True
        if colors is None:
            for label in labels:
                lines.append((label, {}))
        elif colors is not None and not is_markers_list:
            for label, color in zip(labels, colors):
                lines.append((label, {'color': color}))
        elif colors is not None and is_markers_list:
            for label, color, marker in zip(labels, colors, markers):
                lines.append((label, {'color': color, 'marker': marker}))

    if reverse_columns:
        lines = lines[::-1]

    # new legend
    ax.legend(
        [create_dummy_line(**l[1]) for l in lines],  # Line handles
        [l[0] for l in lines],  # Line titles
        loc=legend_loc,
        labelspacing=0.2,  # The vertical space between the legend entries.
        prop={'size': fontsize},
        bbox_to_anchor=bbox_to_anchor,
        framealpha=framealpha,
        handlelength=handlelength,
        title=legend_title,
        ncol=ncol,
        numpoints=numpoints,
        title_fontsize=fontsize,
        facecolor=facecolor,
    )

    set_legend_colors(ax,
                      text_weight=text_weight,
                      fontsize=fontsize,
                      **kwargs)

    ax.get_legend().get_frame().set_linewidth(0.0)


def set_legend_with_stats_table(stats_table: pd.DataFrame,
                                ax: plt.Subplot,
                                colors: List[str],
                                fontsize: int = 10,
                                legend_loc: Optional[str] = 'upper left',
                                bbox_to_anchor: Tuple[float, float] = None,
                                handlelength: float = 1.0,
                                **kwargs
                                ) -> None:
    """
    convert summary table to string and iterate over lines
    """
    # stats_str = stats.to_string(index_names=False)
    stats_str = dfs.df_all_to_str(df=stats_table)
    legend_title = stats_str.splitlines()[0]  # column names will be titles
    lines = []
    for line, color in zip(stats_str.splitlines()[2:], colors):  # var data is from 2-nd line
        lines.append((line, {'color': color}))
    set_legend(ax=ax,
               legend_title=legend_title,
               lines=lines,
               legend_loc=legend_loc,
               fontsize=fontsize,
               handlelength=handlelength,
               bbox_to_anchor=bbox_to_anchor,
               **kwargs)


def set_linestyles(ax: plt.Subplot,
                   linestyles: List[str]
                   ) -> None:
    leg = ax.get_legend()
    leg_lines = leg.get_lines()
    for idx, linestyle in enumerate(linestyles):
        ax.lines[idx].set_linestyle(linestyle)
        leg_lines[idx].set_linestyle(linestyle)


class LegendStats(Enum):
    NONE = 1
    LAST = 2
    AVG = 3
    AVG_LAST = 4
    AVG_STD = 5
    AVG_STD_SKEW_KURT = 6
    AVG_STD_LAST = 7
    AVG_NONNAN_LAST = 8
    NONZERO_AVG_LAST = 81
    NONZERO_AVG_STD_LAST = 82
    MEDIAN_NONNAN_LAST = 9
    AVG_MEDIAN_STD_NONNAN_LAST = 10
    AVG_LAST_SCORE = 11
    AVG_STD_LAST_SCORE = 12
    FIRST_LAST = 13
    FIRST_LAST_NON_ZERO = 14
    FIRST_AVG_LAST = 15
    FIRST_MEDIAN_LAST = 16
    FIRST_AVG_LAST_SHORT = 17
    AVG_STD_MISSING_ZERO = 18
    MISSING_AVG_LAST = 19
    TOTAL = 20
    MEDIAN = 21
    MEDIAN_MAD = 22
    TSTAT = 23
    AVG_STD_TSTAT = 24
    LAST_NONNAN = 25
    AVG_MIN_MAX_LAST = 26
    FIRST_MIN_MAX_LAST = 27


def get_legend_lines(data: Union[pd.DataFrame, pd.Series],
                     legend_stats: LegendStats = LegendStats.NONE,
                     var_format: str = '{:.0f}',
                     tstat_format: str = '{:,.2f}',
                     nan_display: float = np.nan,  # or zero
                     **kwargs
                     ) -> List[str]:

    data = data.copy()
    data = data.replace([np.inf, -np.inf], np.nan)
    if isinstance(data, pd.Series):
        data = data.to_frame()

    if np.any(data.columns.duplicated()):
        print(data)
        raise ValueError(f"dataframe with dublicated columns not supported:\n{print(data.columns)}")

    if legend_stats == LegendStats.NONE:
        legend_lines = data.columns.to_list()

    elif legend_stats == LegendStats.LAST:
        legend_lines = []
        for column in data.columns:
            legend_lines.append(f"{column}: last={var_format.format(data[column].iloc[-1])}")

    elif legend_stats == LegendStats.LAST_NONNAN:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                last = nan_display
            else:
                last = data_column.dropna().iloc[-1]
            legend_lines.append(f"{column} = {var_format.format(last)}")

    elif legend_stats == LegendStats.AVG:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = nan_display
            else:
                avg = np.nanmean(data_column)
            legend_lines.append(f"{column}: avg={var_format.format(avg)}")

    elif legend_stats == LegendStats.MEDIAN:  # specific for fx amm
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                median = nan_display
            else:
                median = np.nanmedian(data_column)
            legend_lines.append(f"{column}: {var_format.format(median)}")

    elif legend_stats == LegendStats.MEDIAN_MAD:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                median = nan_display
                mad = nan_display
            else:
                median = np.nanmedian(data_column)
                mad = stats.median_abs_deviation(data_column.dropna().to_numpy(), scale='normal')
            legend_lines.append(f"{column}: median={var_format.format(median)}, mad={var_format.format(mad)}")

    elif legend_stats == LegendStats.AVG_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = nan_display
                last = np.nan
            else:
                avg = np.nanmean(data_column)
                last = data_column.iloc[-1]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.AVG_STD:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = np.nan
                std = np.nan
            else:
                avg = np.nanmean(data_column)
                std = np.nanstd(data_column, ddof=1)
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, std={var_format.format(std)}")

    elif legend_stats == LegendStats.AVG_STD_SKEW_KURT:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg, std, skw, krt = np.nan, np.nan, np.nan, np.nan
            else:
                data_column = data_column.dropna()
                avg = np.mean(data_column)
                std = np.std(data_column, ddof=1)
                skw = skew(data_column, nan_policy='omit')
                krt = kurtosis(data_column, nan_policy='omit')

            legend_lines.append(f"{column}: avg={var_format.format(avg)}, std={var_format.format(std)}, "
                                f"skew={'{:.2f}'.format(skw)}, kurtosis={'{:.2f}'.format(krt)}")

    elif legend_stats == LegendStats.AVG_STD_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg, std, last = nan_display, nan_display, nan_display
            else:
                avg = np.nanmean(data_column)
                std = np.nanstd(data_column, ddof=1)
                last = data_column.iloc[-1]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, "
                                f"std={var_format.format(std)}, "
                                f"last={var_format.format(last)}")

    elif legend_stats == LegendStats.TSTAT:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                tstat = nan_display
            else:
                avg = np.nanmean(data_column)
                std = np.nanstd(data_column, ddof=1)
                tstat = avg / std
            legend_lines.append(f"{column}: t-stat={tstat_format.format(tstat)}")

    elif legend_stats == LegendStats.AVG_STD_TSTAT:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg, std, tstat = nan_display, nan_display, nan_display
            else:
                avg = np.nanmean(data_column)
                std = np.nanstd(data_column, ddof=1)
                tstat = avg / std
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, "
                                f"std={var_format.format(std)}, "
                                f"t-stat={tstat_format.format(tstat)}")

    elif legend_stats == LegendStats.AVG_NONNAN_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = nan_display
                last = nan_display
            else:
                avg = np.nanmean(data_column)
                last = data_column.dropna().iloc[-1]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.NONZERO_AVG_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column].replace({0.0: np.nan})
            if np.all(np.isnan(data_column)):
                avg = nan_display
                last = nan_display
            else:
                avg = np.nanmean(data_column)
                last = data_column.dropna().iloc[-1]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.NONZERO_AVG_STD_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column].replace({0.0: np.nan})
            if np.all(np.isnan(data_column)):
                avg, std, last = nan_display, nan_display, nan_display
            else:
                avg = np.nanmean(data_column)
                std = np.nanstd(data_column)
                last = data_column.dropna().iloc[-1]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, std={var_format.format(std)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.MEDIAN_NONNAN_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                med = nan_display
                last = nan_display
            else:
                med = np.nanmedian(data_column)
                last = data_column.dropna().iloc[-1]
            legend_lines.append(f"{column}: median={var_format.format(med)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.AVG_MEDIAN_STD_NONNAN_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = nan_display
                med = nan_display
                std = nan_display
                last = nan_display
            else:
                avg = np.nanmean(data_column)
                med = np.nanmedian(data_column)
                std = np.nanstd(data_column)
                last = data_column.dropna().iloc[-1]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, median={var_format.format(med)}, std={var_format.format(std)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.AVG_LAST_SCORE:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = nan_display
                last = nan_display
                score = nan_display
            else:
                data_np = data_column.dropna()
                avg = np.nanmean(data_np)
                last = data_np.iloc[-1]
                score = 0.01 * stats.percentileofscore(a=data_np, score=last, kind='rank')

            legend_lines.append(f"{column}: avg={var_format.format(avg)},"
                                f" last={var_format.format(last)}, "
                                f"last score={'{:.0%}'.format(score)}")

    elif legend_stats == LegendStats.AVG_STD_LAST_SCORE:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = nan_display
                std = nan_display
                last = nan_display
                score = nan_display
            else:
                data_np = data_column.dropna()
                avg = np.nanmean(data_np)
                std = np.nanstd(data_np)
                last = data_np.iloc[-1]
                score = 0.01 * stats.percentileofscore(a=data_np, score=last, kind='rank')

            legend_lines.append(f"{column}: avg={var_format.format(avg)},"
                                f" std={var_format.format(std)}, "
                                f" last={var_format.format(last)}, "
                                f"last score={'{:.0%}'.format(score)}")

    elif legend_stats == LegendStats.FIRST_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                first = nan_display
                last = nan_display
            else:
                nonnan_data = data_column.dropna()
                first = nonnan_data.iloc[0]
                last = nonnan_data.iloc[-1]
            legend_lines.append(f"{column}: first={var_format.format(first)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.FIRST_LAST_NON_ZERO:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                first = nan_display
                last = nan_display
            else:
                data_column = data_column.replace({0.0: np.nan})
                nonnan_data = data_column.dropna()
                first = nonnan_data.iloc[0]
                last = nonnan_data.iloc[-1]
            legend_lines.append(f"{column}: first={var_format.format(first)}, last={var_format.format(last)}")

    elif legend_stats in [LegendStats.FIRST_AVG_LAST, LegendStats.FIRST_AVG_LAST_SHORT]:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                first = nan_display
                avg = nan_display
                last = nan_display
            else:
                nonnan_data = data_column.dropna()
                first = nonnan_data.iloc[0]
                avg = np.nanmean(nonnan_data)
                last = nonnan_data.iloc[-1]
            if legend_stats == LegendStats.FIRST_AVG_LAST_SHORT:
                legend_lines.append(f"{column}: [{var_format.format(first)}, "
                                    f"{var_format.format(avg)}, "
                                    f"{var_format.format(last)}]")
            else:
                legend_lines.append(f"{column}: first={var_format.format(first)}, "
                                    f"avg={var_format.format(avg)}, "
                                    f"last={var_format.format(last)}")

    elif legend_stats == LegendStats.FIRST_MEDIAN_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                first = nan_display
                med = nan_display
                last = nan_display
            else:
                nonnan_data = data_column.dropna()
                first = nonnan_data.iloc[0]
                med = np.nanmedian(nonnan_data)
                last = nonnan_data.iloc[-1]
            legend_lines.append(f"{column}: first={var_format.format(first)}, "
                                f"median={var_format.format(med)}, "
                                f"last={var_format.format(last)}")

    elif legend_stats == LegendStats.AVG_STD_MISSING_ZERO:
        legend_lines = []
        missing_ratio, zeros_ratio = dfo.compute_nans_zeros_ratio_after_first_non_nan(df=data)
        for idx, column in enumerate(data.columns):
            column_data = data[column]
            if np.all(np.isnan(column_data)):
                avg = nan_display
                std = nan_display
                # missing = 1.0
                zeros = nan_display
            else:
                avg = np.nanmean(column_data)
                std = np.nanstd(column_data, ddof=1)
                # missing = missing_ratio[idx]
                zeros = zeros_ratio[idx]
            legend_lines.append(f"{column}: avg={var_format.format(avg)}, "
                                f"std={var_format.format(std)}, "
                                # f"missing%={'{:0.2%}'.format(missing)}, "
                                f"missing%={'{:0.2%}'.format(zeros)}")

    elif legend_stats == LegendStats.MISSING_AVG_LAST:
        legend_lines = []
        missing_ratio, zeros_ratio = dfo.compute_nans_zeros_ratio_after_first_non_nan(df=data)
        for idx, column in enumerate(data.columns):
            column_data = data[column]
            if np.all(np.isnan(column_data)):
                avg = nan_display
                last = nan_display
                zeros = nan_display
            else:
                nonnan_data = column_data.dropna()
                avg = np.nanmean(nonnan_data)
                last = nonnan_data.iloc[-1]
                zeros = zeros_ratio[idx]
            legend_lines.append(f"{column}: "
                                f"missing%={'{:0.2%}'.format(zeros)}, "
                                f"avg={var_format.format(avg)}, "
                                f"last={var_format.format(last)}")

    elif legend_stats == LegendStats.TOTAL:
        legend_lines = []
        for column in data.columns:
            column_data = data[column]
            if np.all(np.isnan(column_data)):
                total = nan_display
            else:
                total = np.nansum(column_data)
            legend_lines.append(f"{column}: total={var_format.format(total)}")
    
    elif legend_stats == LegendStats.AVG_MIN_MAX_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                avg = np.nan
                min = np.nan
                max = np.nan
                last = np.nan
            else:
                nonnan_data = data_column.dropna()
                avg = np.nanmean(nonnan_data)
                min = np.min(nonnan_data)
                max = np.max(nonnan_data)
                last = nonnan_data.iloc[-1]

            legend_lines.append(f"{column}: avg={var_format.format(avg)}, min={var_format.format(min)}, "
                                f"max={var_format.format(max)}, last={var_format.format(last)}")

    elif legend_stats == LegendStats.FIRST_MIN_MAX_LAST:
        legend_lines = []
        for column in data.columns:
            data_column = data[column]
            if np.all(np.isnan(data_column)):
                first = np.nan
                min = np.nan
                max = np.nan
                last = np.nan
            else:
                nonnan_data = data_column.dropna()
                first = nonnan_data.iloc[0]
                min = np.min(nonnan_data)
                max = np.max(nonnan_data)
                last = nonnan_data.iloc[-1]

            legend_lines.append(f"{column}: first={var_format.format(first)}, min={var_format.format(min)}, "
                                f"max={var_format.format(max)}, last={var_format.format(last)}")
        
    else:
        raise TypeError(f"{legend_stats} not implemented")

    return legend_lines


def get_n_colors(n: int,
                 first_color_fixed: bool = False,
                 last_color_fixed: bool = False,
                 fixed_color: str = 'orangered',
                 type: str = 'soft',
                 is_fixed_n_colors: bool = True,
                 **kwargs
                 ) -> List[str]:
    if is_fixed_n_colors:
        colors = get_n_fixed_colors(n=n,
                                    first_color_fixed=first_color_fixed,
                                    last_color_fixed=last_color_fixed,
                                    fixed_color=fixed_color,  # oragne
                                    type=type,
                                    **kwargs)
    else:
        colors = get_n_cmap_colors(n=n, type=type)
        """
        colors = get_n_mlt_colors(n=n,
                                  first_color_fixed=first_color_fixed,
                                  last_color_fixed=last_color_fixed,
                                  fixed_color=fixed_color,  # oragne
                                  type=type,
                                  **kwargs)
        """
    return colors


def get_n_fixed_colors(n: int,
                       first_color_fixed: bool = False,
                       last_color_fixed: bool = False,
                       fixed_color: str = mcolors['darkorange'],
                       type: str = 'soft',
                       is_hex: bool = False,
                       **kwargs
                       ) -> List[str]:
    """
    colors as '#00284A', '#008B75', ...
    """
    if first_color_fixed:
        colors = [fixed_color]
        n -= 1
    else:
        colors = []

    n_clip = np.minimum(n, len(FixedColors))
    n_to_extend = n-n_clip
    for idx, color in enumerate(FixedColors):
        if idx == n_clip:
            break
        colors.append(color.value)

    if n_to_extend > 0:
        colors.extend(get_n_cmap_colors(n=n_to_extend, type=type, is_hex=is_hex))
    if last_color_fixed:
        colors[-1] = fixed_color
    return colors


def get_n_mlt_colors(n: int,
                     first_color_fixed: bool = False,
                     last_color_fixed: bool = False,
                     fixed_color: str = '#EF5A13',  # orange
                     **kwargs
                     ) -> List[str]:
    """
    colors as '#00284A', '#008B75', ...
    """
    if first_color_fixed:
        colors = [fixed_color]
        n -= 1
    else:
        colors = []
    # colors.extend(get_n_cmap_colors(n=n, type=type, is_hex=is_hex))
    colors.extend(get_n_sns_colors(n=n))

    if last_color_fixed:
        colors[-1] = fixed_color
    return colors


def get_n_hatch(n: int) -> List[str]:
    all_hatch = ["//", "\ \\", "-", "+", "x", "o", "O", ".", "*","|"]
    return all_hatch[:n]


def get_n_markers(n: int) -> List[str]:
    all_markers = ["o", "v",  "D", "p",  "^", "<", "s", ">"]
    return all_markers[:n]


def get_n_cmap_colors(n: int,
                      type: str = 'soft',
                      first_color_fixed: bool = False,
                      is_hex: bool = True
                      ) -> List[str]:
    if n == 1:
        colors = ['gray']
    else:
        cmap = rand_cmap(n, first_color_fixed=first_color_fixed, type=type)
        colors = [cmap(i) for i in range(n)]
        if is_hex:
            colors = [rgb2hex(color) for color in colors]
    return colors


def rand_cmap(nlabels: int,
              type: str = 'bright',
              first_color_fixed: bool = True,
              last_color_black: bool = False,
              verbose: bool = False
              ) -> Optional[LinearSegmentedColormap]:
    """
    Creates a random colormap to be used together with matplotlib. Useful for segmentation tasks
    :param nlabels: Number of labels (size of colormap)
    :param type: 'bright' for strong colors, 'soft' for pastel colors
    :param first_color_fixed: Option to use first color as black, True or False
    :param last_color_black: Option to use last color as black, True or False
    :param verbose: Prints the number of labels and shows the colormap. True or False
    :return: colormap for matplotlib
    """
    np.random.seed(1)
    if type not in ('bright', 'soft'):
        print ('Please choose "bright" or "soft" for type')
        return

    if verbose:
        print('Number of labels: ' + str(nlabels))

    # Generate color map for bright colors, based on hsv
    if type == 'bright':
        randHSVcolors = [(np.random.uniform(low=0.0, high=1),
                          np.random.uniform(low=0.2, high=1),
                          np.random.uniform(low=0.9, high=1)) for _ in range(nlabels)]

        # Convert HSV list to RGB
        rand_rg_bcolors = []
        for HSVcolor in randHSVcolors:
            rand_rg_bcolors.append(colorsys.hsv_to_rgb(HSVcolor[0], HSVcolor[1], HSVcolor[2]))

        if first_color_fixed:
            rand_rg_bcolors[0] = [0, 0, 0]

        if last_color_black:
            rand_rg_bcolors[-1] = [0, 0, 0]

        random_colormap = LinearSegmentedColormap.from_list('new_map', rand_rg_bcolors, N=nlabels)

    # Generate soft pastel colors, by limiting the RGB spectrum
    elif type == 'soft':
        low = 0.6
        high = 0.95
        rand_rg_bcolors = [(np.random.uniform(low=low, high=high),
                          np.random.uniform(low=low, high=high),
                          np.random.uniform(low=low, high=high)) for _ in range(nlabels)]

        if first_color_fixed:
            rand_rg_bcolors[0] = [0, 0, 0]

        if last_color_black:
            rand_rg_bcolors[-1] = [0, 0, 0]
        random_colormap = LinearSegmentedColormap.from_list('new_map', rand_rg_bcolors, N=nlabels)

    else:
        raise ValueError(f"unknown type = {type}")

    return random_colormap


def get_cmap_colors(n: int, cmap: str = 'RdYlGn') -> List[str]:
    """
    Returns a list of matplotlib cmap colors
    """
    cmap = plt.cm.get_cmap(cmap, n)
    colors = [cmap(n_) for n_ in range(n)]
    return colors


def get_n_sns_colors(n: int,
                     palette:str = 'bright',
                     desat: Optional[float] = None,
                     as_cmap: bool = False,
                     **kwargs) -> List[str]:
    return sns.color_palette(palette=palette, n_colors=n, desat=desat, as_cmap=as_cmap)


def compute_heatmap_colors(a: np.ndarray,
                           axis: int = None,
                           cmap: str = 'RdYlGn',
                           alpha: float = 0.75,  # color scaler
                           exclude_max: bool = False,
                           **kwargs
                           ) -> List[Tuple[float, float, float]]:
    """
    for usage in meta
    axis defines maps by total, columns, rows Literal[None, 0, 1] = None
    """
    lower = np.nanmin(a, axis=axis, keepdims=True)
    upper = np.nanmax(a, axis=axis, keepdims=True)
    if exclude_max:  # avoid too extreeme values
        max_idx = a.argmax()
        a[max_idx] = np.nan
        upper = np.nanmax(a, axis=axis, keepdims=True)
        a[max_idx] = upper

    diffs = upper - lower
    scaler = np.reciprocal(diffs, where=np.greater(diffs, 0.0))
    cond = np.logical_and(np.isfinite(scaler), np.isfinite(scaler))
    z = alpha*np.where(cond, scaler * (a - lower), np.nan)

    if a.ndim == 1:
        # colorise = [(1 - x, x, left) if np.isnan(x) == False else (1, 1, 1) for x in z]
        colorise = [x for x in plt.get_cmap(cmap)(z)]

    elif a.ndim == 2:  # list of lists
        colors = plt.get_cmap(cmap)(z)
        colorise = []
        for idx, row in enumerate(z):
            colorise.append(colors[idx])
    else:
        raise ValueError(f"unsupported {a.ndim}")

    return colorise


def get_data_group_colors(df: pd.DataFrame,
                          x: str,
                          y: str,
                          hue: str = None,
                          is_bullish: bool = True
                          ) -> List[Tuple[float, float, float]]:

    avg = df.groupby(x, sort=False).mean().sort_values(y)[y]
    colorise = compute_heatmap_colors(a=avg.to_numpy())
    if not is_bullish:
        colorise = colorise[::-1]

    if hue is not None:
        n_hue = len(df[hue].unique())
        # colorise = [elem for _ in range(n_hue) for elem in colorise]  # extend colors n times
        colorise = [elem for elem in colorise for _ in range(n_hue) ]  # extend colors n times
    return colorise


def add_scatter_points(ax: plt.Subplot,
                       label_x_y: Union[Dict[str, Tuple[float, float]], List[Tuple[float, float]]],
                       fontsize: int = 12,
                       color: str = 'steelblue',
                       colors: List[str] = None,
                       linewidth: int = 3,
                       marker: str = '*',
                       size: int = 3,
                       **kwargs
                       ) -> None:
    """
    add scatter points to ax
    """
    if colors is None:
        if isinstance(label_x_y, dict):
            colors = len(label_x_y.keys()) * [color]
        else:
            colors = len(label_x_y) * [color]
    if isinstance(label_x_y, dict):
        for (key, (x, y)), color in zip(label_x_y.items(), colors):
            ax.annotate(key, xy=(x, y), xytext=(2, 2), color=color,
                        textcoords='offset points', ha='left', va='bottom', fontsize=fontsize)
            ax.scatter(x=x, y=y, marker=marker, color=color, s=size, linewidth=linewidth)
    else:
        for (x, y), color in zip(label_x_y, colors):
            ax.scatter(x=x, y=y, marker=marker, color=color, s=size, linewidth=linewidth)


def calc_table_height(num_rows: int,
                      row_height: float = ROW_HIGHT,
                      first_row_height: float = None,
                      scale: float = 0.225
                      ) -> float:
    """
    core function for sizing of table and heatmaplot
    """
    if first_row_height is None:
        first_row_height = row_height
    height = (scale*num_rows*row_height + scale*first_row_height)
    return height


def calc_table_width(num_col: int,
                     column_width: float = COLUMN_WIDTH,
                     first_column_width: Optional[float] = FIRST_COLUMN_WIDTH,
                     scale: float = 0.225
                     ) -> float:
    """
    core function for sizing of table and heatmaplot
    """
    if first_column_width is None:
        first_column_width = first_column_width
    width = (scale*num_col*column_width + scale*first_column_width)
    return width


def calc_df_table_size(df: pd.DataFrame,
                       min_rows: Optional[int] = None,
                       min_cols: Optional[int] = None,
                       scale_rows: float = 0.225,
                       scale_cols: float = 0.225
                       ) -> Tuple[float, float, int, int]:
    """
    calc optimal table size
    """
    if min_rows is not None:
        num_rows = np.maximum(len(df.index), min_rows)
    else:
        num_rows = len(df.index)
    if min_cols is not None:
        num_cols = np.maximum(len(df.columns), min_cols)
    else:
        num_cols = len(df.columns)
    width = calc_table_width(num_col=num_cols, scale=scale_cols)
    height = calc_table_height(num_rows=num_rows, scale=scale_rows)
    return width, height, num_cols, num_rows


def get_df_table_size(df: pd.DataFrame,
                      min_rows: Optional[int] = None,
                      min_cols: Optional[int] = None,
                      scale_rows: float = 0.225,
                      scale_cols: float = 0.225
                      ) -> Tuple[float, float]:
    """
    calc optimal table size
    """
    width, height, num_cols, num_rows = calc_df_table_size(df=df,
                                                           min_rows=min_rows,
                                                           min_cols=min_cols,
                                                           scale_rows=scale_rows,
                                                           scale_cols=scale_cols)
    return width, height


def reset_xticks(ax: plt.Axes, data: np.ndarray, nbins: int = 20, var_format: str = '{:.2%}') -> None:
    """
    useful for barplot to reduce the number of x ticks
    """
    current_ticks = ax.get_xticks()
    ax.set_xticks(np.linspace(current_ticks[0], current_ticks[-1], nbins))
    x_datalables = np.linspace(data[0], data[-1], nbins)
    ax.set_xticklabels([var_format.format(x) for x in x_datalables], rotation=0, fontsize=10)


class UnitTests(Enum):
    DUMMY_LINE = 1
    LEGEND_LINES = 2
    CMAP_COLORS = 3
    SNS_COLORS = 4
    HEATMAP_COLORS = 5
    GET_COLORS = 6


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.DUMMY_LINE:
        print(create_dummy_line())

    elif unit_test == UnitTests.LEGEND_LINES:
        from qis.test_data import load_etf_data
        prices = load_etf_data().dropna()

        for legend_stats in LegendStats:
            legend_lines = get_legend_lines(data=prices, legend_stats=legend_stats)
            print(legend_lines)

    elif unit_test == UnitTests.CMAP_COLORS:
        cmap_colors = get_cmap_colors(n=100)
        print(cmap_colors)

    elif unit_test == UnitTests.SNS_COLORS:
        cmap_colors = get_n_sns_colors(n=3)
        print(cmap_colors)

    elif unit_test == UnitTests.HEATMAP_COLORS:
        data = np.array([1.0, 2.0, 3.0])
        print(data.ndim)
        heatmap_colors = compute_heatmap_colors(a=data)
        print(heatmap_colors)

        data = np.array([[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]])
        print(data.ndim)
        heatmap_colors = compute_heatmap_colors(a=data)
        print(heatmap_colors)

    elif unit_test == UnitTests.GET_COLORS:
        n_colors = get_n_colors(n=10)
        print(n_colors)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.GET_COLORS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




from qis.portfolio.reports.config import (FactsheetConfig,
                                          FACTSHEET_CONFIG_DAILY_DATA_LONG_PERIOD,
                                          FACTSHEET_CONFIG_DAILY_DATA_SHORT_PERIOD,
                                          FACTSHEET_CONFIG_MONTHLY_DATA_LONG_PERIOD,
                                          FACTSHEET_CONFIG_MONTHLY_DATA_SHORT_PERIOD,
                                          fetch_factsheet_config_kwargs,
                                          fetch_default_perf_params,
                                          fetch_default_report_kwargs)

from qis.portfolio.reports.brinson_attribution import (compute_brinson_attribution_table,
                                                       plot_brinson_attribution_table)

from qis.portfolio.backtester import (backtest_model_portfolio, backtest_rebalanced_portfolio)

from qis.portfolio.portfolio_data import PortfolioData, PortfolioInput, AttributionMetric

from qis.portfolio.reports.multi_assets_factsheet import (MultiAssetsReport, generate_multi_asset_factsheet)

from qis.portfolio.reports.strategy_factsheet import generate_strategy_factsheet

from qis.portfolio.reports.strategy_benchmark_factsheet import (generate_strategy_benchmark_factsheet_plt,
                                                                generate_strategy_benchmark_active_perf_plt)

from qis.portfolio.multi_portfolio_data import MultiPortfolioData

from qis.portfolio.reports.multi_strategy_factsheet import generate_multi_portfolio_factsheet

# disable requirements for pyblogs
# from qis.portfolio.reports.multi_strategy_factseet_pybloqs import generate_multi_portfolio_factsheet_with_pyblogs



"""
scatter plot core
"""
# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import scipy.stats as stats
from statsmodels import api as sm
from typing import Union, List, Tuple, Optional
from enum import Enum

# qis
import qis.plots as qp
import qis.utils as qu


def plot_scatter(df: pd.DataFrame,
                 x: str = None,
                 y: str = None,
                 hue: str = None,
                 xlabel: Union[str, bool, None] = True,
                 ylabel: Union[str, bool, None] = True,
                 title: Optional[str] = None,
                 annotation_labels: List[str] = None,
                 annotation_colors: List[str] = None,
                 annotation_color: Optional[str] = 'red',
                 add_universe_model_label: bool = True,
                 add_universe_model_prediction: bool = False,
                 add_universe_model_ci: bool = False,
                 add_hue_model_label: Optional[bool] = None,  # add hue eqs for data with hue
                 ci: Optional[int] = None,
                 order: int = 2,  # regression order
                 full_sample_order: Optional[int] = 2,  # full sample order can be different
                 fit_intercept: bool = True,
                 full_sample_color: str = 'grey',
                 colors: List[str] = None,
                 xvar_format: str = '{:.0%}',
                 yvar_format: str = '{:.0%}',
                 x_limits: Tuple[Optional[float], Optional[float]] = None,
                 y_limits: Tuple[Optional[float], Optional[float]] = None,
                 xticks: List[str] = None,
                 fontsize: int = 10,
                 linewidth: float = 1.5,
                 markersize: int = 4,
                 full_sample_label: str = 'Full sample: ',
                 add_45line: bool = False,
                 r2_only: bool = False,
                 legend_loc: str = 'upper left',
                 value_name: str = 'value_name',
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> plt.Figure:
    """
    x-y scatter of df
    """
    df = df.copy().dropna()

    if x is None:
        if len(df.columns) == 2 or (len(df.columns) == 3 and hue is not None):
            x = df.columns[0]
        else:
            raise ValueError(f"x_column is not defined for more than on columns")
    if y is None:
        if len(df.columns) == 2 or (len(df.columns) == 3 and hue is not None):  # x and y
            y = df.columns[1]
        else:  # melting to column value_name with hue = all columns ba t x
            hue = 'hue'
            y = value_name
            df = pd.melt(df, id_vars=[x], value_vars=df.columns.drop(x), var_name=hue,
                         value_name=value_name)

    if hue is not None and add_hue_model_label is None:  # override to true unless false
        add_hue_model_label = True

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
    else:
        fig = None

    estimated_reg_models = {}
    if hue is not None:
        if colors is None:
            colors = qp.get_n_sns_colors(n=len(df[hue].unique()), **kwargs)
        palette = colors

        hue_ids = df[hue].unique()
        for idx, hue_id in enumerate(hue_ids):
            # estimate model equation
            data_hue = df[df[hue] == hue_id].sort_values(by=x)
            x_ = data_hue[x].to_numpy()
            y_ = data_hue[y].to_numpy()
            x1 = qu.get_ols_x(x=x_, order=order, fit_intercept=fit_intercept)
            reg_model = sm.OLS(y_, x1).fit()
            estimated_reg_models[hue_id] = reg_model

            # plot data points
            sns.scatterplot(x=x, y=y, data=data_hue, color=palette[idx], s=markersize, ax=ax)

            if order > 0:  # plot prediction
                if ci is not None: # not possible to control reg equation in regplot, only use for ci
                    sns.regplot(x=x, y=y, data=data_hue,
                                ci=ci,
                                order=order, truncate=True,
                                color=palette[idx],
                                scatter_kws={'s': markersize},
                                line_kws={'linewidth': linewidth},
                                ax=ax)
                else:
                    prediction = reg_model.predict(x1)
                    ax.plot(x_, prediction, color=palette[idx], lw=linewidth, linestyle='-')

    else:
        if full_sample_order is None:
            pass
        elif full_sample_order == 0:  # just scatter plot
            sns.scatterplot(x=x, y=y, data=df,
                            # ci=ci,
                            s=markersize, color=full_sample_color, ax=ax)
        else:  # regplot add scatter and ml lines even if order is == 0
            sns.regplot(x=x, y=y, data=df, ci=ci, order=full_sample_order, color=full_sample_color,
                        scatter_kws={'s': markersize},
                        line_kws={'linewidth': linewidth}, ax=ax)

    # add ml equations to labels
    legend_labels = []
    legend_colors = []
    if full_sample_order is not None:
        if (add_universe_model_prediction or add_universe_model_label or add_universe_model_ci) and full_sample_order > 0:
            xy = df[[x, y]].sort_values(by=x)
            x_ = xy[x].to_numpy()
            y_ = xy[y].to_numpy()
            x1 = qu.get_ols_x(x=x_, order=full_sample_order, fit_intercept=fit_intercept)
            reg_model = sm.OLS(y_, x1).fit()

            if add_universe_model_prediction:
                prediction = reg_model.predict(x1)
                ax.plot(x, prediction, color=full_sample_color, lw=linewidth, linestyle='--')

            if add_universe_model_ci:
                y_model = reg_model.predict(x1)
                ci = calc_ci(x=x, y=y, y_model=y_model)
                # ax.fill_between(x, y + ci, y - ci, color="None", linestyle="--")
                ax.plot(x, y_model - ci, "--", color="0.5")
                ax.plot(x, y_model + ci, "--", color="0.5")

            if add_universe_model_label:
                text_str = f"{full_sample_label} " \
                           f"{qu.reg_model_params_to_str(reg_model=reg_model, order=full_sample_order, r2_only=False, fit_intercept=fit_intercept, **kwargs)}"
                legend_labels.append(text_str)
                legend_colors.append(full_sample_color)

    # add colors for annotation labels
    df['color'] = full_sample_color
    if hue is not None :
        hue_ids = df[hue].unique()
        for color, hue_id in zip(colors, hue_ids):
            df.loc[df[hue] == hue_id, 'color'] = 'red'  # ad color for hue
            if order > 0:
                if add_hue_model_label:
                    reg_model = estimated_reg_models[hue_id]
                    text_str = (f"{hue_id}: " 
                                f"{qu.reg_model_params_to_str(reg_model=reg_model, order=order, r2_only=r2_only, fit_intercept=fit_intercept, **kwargs)}")
                else:
                    text_str = hue_id
                legend_labels.append(text_str)
            else:
                legend_labels.append(hue_id)
            legend_colors.append(color)

    elif hue is not None and order == 0:
        legend_labels = df[hue].unique()
        legend_colors = colors

    # add labels
    if annotation_labels is not None:
        if annotation_colors is not None:
            colors = annotation_colors
        elif annotation_color is not None:
            colors = len(df.index) * [annotation_color]
        else:
            colors = df['color']
        for label, x_, y_, color in zip(annotation_labels, df[x], df[y], colors):
            ax.annotate(label,
                        xy=(x_, y_), xytext=(1, 1),
                        textcoords='offset points', ha='left', va='bottom',
                        color=color,
                        fontsize=fontsize)
            if label != '':
                ax.scatter(x=x_, y=y_, c=color, s=20)

    if add_45line:  # make equal:
        ymin, ymax = ax.get_ylim()
        xmin, xmax = ax.get_xlim()
        min = ymin if ymin < xmin else xmin
        max = ymax if ymax > xmax else xmax
        ax.set_xlim([min, max])
        ax.set_ylim([min, max])
        x = np.linspace(*ax.get_xlim())
        ax.plot(x, x, color='black', lw=1, linestyle='--')

    if x_limits is not None:
        qp.set_x_limits(ax=ax, x_limits=x_limits)
    if y_limits is not None:
        qp.set_y_limits(ax=ax, y_limits=y_limits)

    if xticks is not None:
        qp.set_ax_tick_labels(ax=ax, x_rotation=0, xticks=df[x].to_numpy(), x_labels=xticks,
                              fontsize=fontsize)
    else:
        qp.set_ax_tick_labels(ax=ax, x_rotation=0, fontsize=fontsize)

    if isinstance(xlabel, bool):
        if xlabel is True:
            xlabel = f"x={x}"
        else:
            xlabel = ''
    if isinstance(ylabel, bool):
        if ylabel is True:
            ylabel = f"y={y}"
        else:
            ylabel = ''
    qp.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)

    qp.set_ax_ticks_format(ax=ax, xvar_format=xvar_format, yvar_format=yvar_format, fontsize=fontsize, **kwargs)

    qp.set_legend(ax=ax,
                  labels=legend_labels,
                  colors=legend_colors,
                  legend_loc=legend_loc,
                  fontsize=fontsize,
                  **kwargs)

    if title is not None:
        qp.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    qp.set_spines(ax=ax, **kwargs)

    return fig


def plot_classification_scatter(df: pd.DataFrame,
                                x: Optional[str] = None,
                                y: Optional[str] = None,
                                hue_name: str = 'hue',
                                num_buckets: Optional[int] = None,
                                bins: np.ndarray = np.array([-3.0, -1.5, 0.0, 1.5, 3.0]),
                                order: int = 1,
                                title: str = None,
                                full_sample_order: Optional[int] = 3,
                                markersize: int = 10,
                                xvar_format: str = '{:.2f}',
                                yvar_format: str = '{:.2f}',
                                fit_intercept: bool = False,
                                ax: plt.Subplot = None,
                                **kwargs
                                ) -> Optional[plt.Figure]:
    """
    add bin classification using x_column
    """
    if x is None:
        if len(df.columns) == 2:
            x = df.columns[0]
        else:
            raise ValueError(f"x_column is not defined for more than on columns")
    if y is None:
        if len(df.columns) == 2:  # x and y
            y = df.columns[1]
        else:
            raise ValueError(f"y_column is not defined for more than on columns")

    df, _ = qu.add_quantile_classification(df=df, x_column=x, hue_name=hue_name, num_buckets=num_buckets, bins=bins)

    fig = plot_scatter(df=df,
                       x=x,
                       y=y,
                       hue=hue_name,
                       fit_intercept=fit_intercept,
                       title=title,
                       order=order,
                       full_sample_order=full_sample_order,
                       markersize=markersize,
                       xvar_format=xvar_format,
                       yvar_format=yvar_format,
                       add_universe_model_ci=False,
                       add_hue_model_label=True,
                       ax=ax,
                       **kwargs)
    return fig


def calc_ci(x: np.ndarray, y: np.ndarray, y_model: np.ndarray) -> np.ndarray:
    n = x.shape[0]
    m = 2
    dof = n - m
    t = stats.t.ppf(0.95, dof)
    #x2 = np.linspace(np.min(x), np.max(x), 100)

    # Estimates of Error in Data/Model
    resid = y - y_model
    # chi2 = np.sum((resid / y_model) ** 2)  # chi-squared; estimates error in data
    # chi2_red = chi2 / dof  # reduced chi-squared; measures goodness of fit
    s_err = np.sqrt(np.sum(resid ** 2) / dof)  # standard deviation of the error

    ci = t * s_err * np.sqrt(1 / n + (x - np.mean(x)) ** 2 / np.sum((x - np.mean(x)) ** 2))
    return ci


def get_random_data(is_random_beta: bool = True,
                    n: int = 10000
                    ) -> pd.DataFrame:

    x = np.random.normal(0.0, 1.0, n)
    eps = np.random.normal(0.0, 1.0, n)
    if is_random_beta:
        beta = np.random.normal(1.0, 1.0, n)*np.square(x)
    else:
        beta = np.ones(n)
    y = beta*x + eps
    df = pd.concat([pd.Series(x, name='x'), pd.Series(y, name='y')], axis=1)
    df = df.sort_values(by='x', axis=0)

    return df


def estimate_classification_scatter(df: pd.DataFrame,
                                    x: Optional[str] = None,
                                    y: Optional[str] = None,
                                    hue_name: str = 'hue',
                                    num_buckets: Optional[int] = None,
                                    bins: np.ndarray = np.array([-3.0, -1.5, 0.0, 1.5, 3.0]),
                                    order: int = 1,
                                    fit_intercept: bool = False,
                                    ) -> pd.Series:
    """
    add bin classification using x_column
    """
    if x is None:
        if len(df.columns) == 2:
            x = df.columns[0]
        else:
            raise ValueError(f"x_column is not defined for more than on columns")
    if y is None:
        if len(df.columns) == 2:  # x and y
            y = df.columns[1]
        else:
            raise ValueError(f"y_column is not defined for more than on columns")

    df, _ = qu.add_quantile_classification(df=df, x_column=x, hue_name=hue_name, num_buckets=num_buckets, bins=bins)

    x_np = df[x].to_numpy()
    x_range = np.linspace(np.min(x_np), np.max(x_np), 200)

    dfs = df.groupby(hue_name)
    y_rpeds = []
    for hue, df in dfs:
        # estimate model equation
        data_hue = df.sort_values(by=x)
        x_ = data_hue[x].to_numpy()
        y_ = data_hue[y].to_numpy()
        x1 = qu.get_ols_x(x=x_, order=order, fit_intercept=fit_intercept)
        reg_model = sm.OLS(y_, x1).fit()
        x_hue = np.extract(np.logical_and(x_range>=np.min(x_), x_range<np.max(x_)), x_range)
        y_hue = reg_model.predict(x_hue)
        y_rpeds.append(pd.Series(y_hue, index=x_hue))
    y_rpeds = pd.concat(y_rpeds).sort_index()
    return y_rpeds


class UnitTests(Enum):
    SCATTER = 1
    CLASSIFICATION_SCATTER = 2
    CLASSIFICATION_REGRESSION = 3


def run_unit_test(unit_test: UnitTests):

    np.random.seed(2)
    df = get_random_data(n=100000)
    print(df)

    if unit_test == UnitTests.SCATTER:
        plot_scatter(df=df)

    elif unit_test == UnitTests.CLASSIFICATION_SCATTER:
        plot_classification_scatter(df=df, x='x', y='y')

    elif unit_test == UnitTests.CLASSIFICATION_REGRESSION:
        y_rpeds = estimate_classification_scatter(df=df, x='x', y='y')
        plot_classification_scatter(df=df, x='x', y='y')
        print(y_rpeds)
        y_rpeds.plot()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CLASSIFICATION_REGRESSION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
various df functions to create scores of df data
"""
import numpy as np
import pandas as pd
from typing import Union, Optional, List


def df_to_cross_sectional_score(df: Union[pd.Series, pd.DataFrame],
                                lower_clip: Optional[float] = -5.0,
                                upper_clip: Optional[float] = 5.0
                                ) -> Union[pd.Series, pd.DataFrame]:
    """
    compute cross sectional score
    """
    if lower_clip is not None or upper_clip is not None:
        df = df.clip(lower=lower_clip, upper=upper_clip)

    if isinstance(df, pd.Series):
        score = (df - np.nanmean(df)) / np.nanstd(df)
    else:
        score = (df - np.nanmean(df, axis=1, keepdims=True)) / np.nanstd(df, axis=1, keepdims=True)
    return score


def df_to_max_score(df: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """
    normalized rows by cross-sectional max: max element = 1.0
    """
    if isinstance(df, pd.Series):
        score = df.divide(np.nanmax(df, axis=0))
    else:
        score = df.divide(np.nanmax(df, axis=1, keepdims=True))
    return score


def compute_aggregate_scores(scores: List[pd.Series],
                             lower_clip: Optional[float] = -5.0,
                             upper_clip: Optional[float] = 5.0,
                             normalize_to_unit_std: bool = True,
                             penalise_nan_values: bool = True
                             ) -> pd.Series:
    """
    aggregate list scores
    nan values are penalised
    normalize_to_unit_std: avg of scores is unit std
    """
    joint = pd.concat(scores, axis=1).clip(lower=lower_clip, upper=upper_clip)
    if penalise_nan_values: # scores with nan are penalised
        n = len(scores)
        if normalize_to_unit_std:
            norm = 1.0 / np.sqrt(n)
        else:
            norm = 1 / n
        joint_avg = norm * np.nansum(joint, axis=1)
    else:
        joint_avg = np.nanmean(joint, axis=1)
    joint_score = pd.Series(joint_avg, index=joint.index).sort_values(ascending=False)
    return joint_score



"""
common python structures operations
"""
import itertools
import pandas as pd
from collections.abc import Iterable
from enum import Enum
from typing import Dict, Tuple, List, Any, NamedTuple, Union, Optional


def list_intersection(list_check: Union[List[Any], pd.Index],
                      list_sample: Union[List[Any], pd.Index]
                      ) -> List[Any]:
    list_out = list(filter(lambda x: x in list_check, list_sample))
    if list_out is None:
        list_out = []
    return list_out


def list_diff(list_check: List[Any],
              list_sample: List[Any]
              ) -> List[Any]:
    list_out = list(filter(lambda x: x not in list_check, list_sample))
    if list_out is None:
        list_out = []
    return list_out


def merge_lists_unique(list1: List[Any],
                       list2: List[Any]
                       ) -> List[Any]:
    list_out = list_intersection(list_check=list2, list_sample=list1)
    list_diffs1 = list_diff(list_check=list_out, list_sample=list1)
    list_diffs2 = list_diff(list_check=list_out, list_sample=list2)
    if len(list_diffs1) > 0:
        list_out.extend(list_diffs1)
    if len(list_diffs2) > 0:
        list_out.extend(list_diffs2)
    return list_out


def assert_list_subset(large_list: List[Any],  # larger list
                       list_sample: List[Any],  # smaller list,
                       is_stop: bool = True,
                       message: str = ''
                       ) -> bool:
    outlyers = list_diff(list_check=large_list, list_sample=list_sample)
    output = True
    if len(outlyers) > 0:
        if is_stop:
            raise ValueError(f"{message}\nnot found {outlyers}")
        else:
            print(f"in assert_list_subset:\n{message}\n{outlyers}")
            output = False
    return output


def list_to_unique_and_dub(lsdata: List) -> Tuple[List, List]:
    """
    find unique and dublicates in list
    """
    unique = set()
    dublicated = []  # can count occurencies if needed
    for x in lsdata:
        if x not in unique:
            unique.add(x)
        else:
            dublicated.append(x)
    unique = list(unique)
    return unique, dublicated


def assert_list_unique(lsdata: List[str]) -> None:
    unique, duplicated = list_to_unique_and_dub(lsdata=lsdata)
    if len(duplicated) > 0:
        raise ValueError(f"list has duplicated elements = {duplicated}")


def move_item_to_first(lsdata: List[Any], item: Any) -> List[Any]:
    out_list = lsdata.copy()
    out_list.remove(item)
    out_list = [item]+out_list
    return out_list


def flatten_dict_tuples(dict_tuples: Dict[str, Union[Any, NamedTuple]]) -> Dict[str, Any]:
    data = {}
    for k, v in dict_tuples.items():
        if isinstance(v, tuple):
            data.update(v._asdict())  # this will create a dict from named tuple
        else:
            data[k] = v
    return data


def split_dict(d: Dict) -> Tuple[Dict, Dict]:
    """
    split dictionary into 2 parts
    """
    n = len(d) // 2
    i = iter(d.items())

    d1 = dict(itertools.islice(i, n))   # grab first n items
    d2 = dict(i)                        # grab the rest

    return d1, d2


def flatten(items: Iterable) -> Any:
    """
    flatten list/items from any nested iterable
    """
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in flatten(x):
                yield sub_x
        else:
            yield x


def to_flat_list(items: Iterable) -> List[Any]:
    return [item for item in flatten(items)]


def update_kwargs(kwargs: Dict[Any, Any],
                  new_kwargs: Optional[Dict[Any, Any]]
                  ) -> Dict[Any, Any]:
    """
    update kwargs with optional kwargs dicts
    """
    local_kwargs = kwargs.copy()
    if new_kwargs is not None and not len(new_kwargs) == 0:
        local_kwargs.update(new_kwargs)
    return local_kwargs


def separate_number_from_string(string: str) -> List[str]:
    """
    given 'A3' get 3
    """
    previous_character = string[0]
    groups = []
    newword = string[0]
    for x, i in enumerate(string[1:]):
        if i.isalpha() and previous_character.isalpha():
            newword += i
        elif i.isnumeric() and previous_character.isnumeric():
            newword += i
        else:
            groups.append(newword)
            newword = i
        previous_character = i
        if x == len(string) - 2:
            groups.append(newword)
            newword = ''

    return groups


class UnitTests(Enum):
    FLATTEN = 1
    LIST = 2
    MERGE = 3
    STRINGS = 4


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.FLATTEN:
        items = [[1, 2], [[3]], 4]
        flat_items = flatten(items)
        [print(item) for item in flat_items]
        print(to_flat_list(items))

    elif unit_test == UnitTests.LIST:
        rows_edge_lines = list(itertools.accumulate(10 * [5]))
        print(rows_edge_lines)

    elif unit_test == UnitTests.MERGE:
        list2 = ['EQ', 'HUI']
        list1 = ['EQ', 'BD', 'STIR', 'FX', 'Energies', 'Metals', 'Ags']
        groups = merge_lists_unique(list1=list1, list2=list2)
        print('groups1')
        print(groups)
        groups = merge_lists_unique(list1=list2, list2=list1)
        print('groups2')
        print(groups)

    elif unit_test == UnitTests.STRINGS:
        string = '123me45you0000me7+33.3'
        this = separate_number_from_string(string)
        print(this)


if __name__ == '__main__':

    unit_test = UnitTests.STRINGS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)


